from pathlib import Path

from dotenv import load_dotenv

# Getting the absolute path of the current file's directory
ABS_PATH = (Path(__file__)).parent

# Getting the absolute path of the project's root directory
PROJECT_DIR = Path(ABS_PATH).parent.parent


def _load_env() -> None:
    dotenv_path = Path(PROJECT_DIR) / "tests" / "integration_tests" / ".env"
    if Path(dotenv_path).exists():
        load_dotenv(dotenv_path)


_load_env()



import os
from typing import Any

from langchain_tests.integration_tests.embeddings import EmbeddingsIntegrationTests

from langchain_ibm import WatsonxEmbeddings

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

URL = "https://us-south.ml.cloud.ibm.com"

MODEL_ID = "ibm/granite-embedding-278m-multilingual"


class TestWatsonxEmbeddingsStandard(EmbeddingsIntegrationTests):
    @property
    def embeddings_class(self) -> type[WatsonxEmbeddings]:
        return WatsonxEmbeddings

    @property
    def embedding_model_params(self) -> dict[str, Any]:
        return {
            "model_id": MODEL_ID,
            "url": URL,
            "apikey": WX_APIKEY,
            "project_id": WX_PROJECT_ID,
        }



import os

import pytest
from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]
from langchain_core.messages import (
    HumanMessage,
    SystemMessage,
)

from langchain_ibm import ChatWatsonx

URL = "https://us-south.ml.cloud.ibm.com"
WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

MODEL = "mistralai/mistral-large"


@pytest.mark.skip("Until not released on us-south, should be skipped.")
class TestChatModelsGateway:
    prompt = "Hello, How are you!"

    messages = (
        ("user", "You are a helpful assistant that translates English to French."),
        (
            "human",
            "Translate this sentence from English to French. I love programming.",
        ),
    )

    system_message = SystemMessage(
        content="You are a helpful assistant which telling "
        "short-info about provided topic."
    )
    human_message = HumanMessage(content="horse")

    def test_chat_model_gateway_init_credentials(self) -> None:
        chat = ChatWatsonx(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        assert isinstance(chat, ChatWatsonx)

        response = chat.invoke(self.prompt)

        assert response
        assert response.content

    def test_chat_model_gateway_init_client(self) -> None:
        credentials = Credentials(url=URL, api_key=WX_APIKEY)
        api_client = APIClient(credentials=credentials, project_id=WX_PROJECT_ID)

        chat = ChatWatsonx(model=MODEL, watsonx_client=api_client)
        assert isinstance(chat, ChatWatsonx)

        response = chat.invoke(self.prompt)

        assert response
        assert response.content

    def test_chat_model_gateway_invoke(self) -> None:
        chat = ChatWatsonx(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.invoke(self.prompt)
        assert response
        assert response.content

        response = chat.invoke(self.messages)
        assert response
        assert response.content

    async def test_chat_model_gateway_ainvoke(self) -> None:
        chat = ChatWatsonx(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = await chat.ainvoke(self.prompt)
        assert response
        assert response.content

        response = await chat.ainvoke(self.messages)
        assert response
        assert response.content

    def test_chat_model_gateway_generate(self) -> None:
        chat = ChatWatsonx(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.generate(messages=[[self.system_message], [self.human_message]])
        assert response
        for generation in response.generations:
            assert generation[0].text

    async def test_chat_model_gateway_agenerate(self) -> None:
        chat = ChatWatsonx(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = await chat.agenerate([[self.system_message], [self.human_message]])
        assert response
        for generation in response.generations:
            assert generation[0].text

    def test_chat_model_gateway_stream(self) -> None:
        chat = ChatWatsonx(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.stream(self.prompt)
        for chunk in response:
            assert isinstance(chunk.content, str)

        response = chat.stream(self.messages)
        for chunk in response:
            assert isinstance(chunk.content, str)

    async def test_chat_model_gateway_astream(self) -> None:
        chat = ChatWatsonx(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.astream(self.prompt)
        async for chunk in response:
            assert isinstance(chunk.content, str)

        response = chat.astream(self.messages)
        async for chunk in response:
            assert isinstance(chunk.content, str)



"""IBM watsonx.ai embeddings wrapper."""

import logging
from typing import Any, cast

from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models.embeddings import (  # type: ignore[import-untyped]
    Embeddings,
)
from ibm_watsonx_ai.gateway import Gateway  # type: ignore[import-untyped]
from langchain_core.embeddings import Embeddings as LangChainEmbeddings
from langchain_core.utils.utils import secret_from_env
from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    SecretStr,
    model_validator,
)
from typing_extensions import Self

from langchain_ibm.utils import (
    async_gateway_error_handler,
    extract_params,
    gateway_error_handler,
    normalize_api_key,
    resolve_watsonx_credentials,
    secret_from_env_multi,
)

logger = logging.getLogger(__name__)


class WatsonxEmbeddings(BaseModel, LangChainEmbeddings):
    """`IBM watsonx.ai` embedding model integration.

    ???+ info "Setup"

        To use, you should have `langchain_ibm` python package installed,
        and the environment variable `WATSONX_API_KEY` set with your API key, or pass
        it as a named parameter `api_key` to the constructor.

        ```bash
        pip install -U langchain-ibm

        # or using uv
        uv add langchain-ibm
        ```

        ```bash
        export WATSONX_API_KEY="your-api-key"
        ```

        !!! deprecated
            `apikey` and `WATSONX_APIKEY` are deprecated and will be removed in
            version `2.0.0`. Use `api_key` and `WATSONX_API_KEY` instead.

    ??? info "Instantiate"

        ```python
        from langchain_ibm import WatsonxEmbeddings

        embeddings = WatsonxEmbeddings(
            model_id="ibm/granite-embedding-278m-multilingual",
            url="https://us-south.ml.cloud.ibm.com",
            project_id="*****",
            # api_key="*****"
        )
        ```

    ??? info "Embed single text"

        ```python
        input_text = "The meaning of life is 42"
        vector = embeddings.embed_query("hello")
        print(vector[:3])
        ```

        ```python
        [-0.0020519258, 0.0147288125, -0.0090887165]
        ```

    ??? info "Embed multiple texts"

        ```python
        vectors = embeddings.embed_documents(["hello", "goodbye"])
        # Showing only the first 3 coordinates
        print(len(vectors))
        print(vectors[0][:3])
        ```

        ```python
        2
        [-0.0020519265, 0.01472881, -0.009088721]
        ```

    ??? info "Async"

        ```python
        await embeddings.aembed_query(input_text)
        print(vector[:3])

        # multiple:
        # await embeddings.aembed_documents(input_texts)
        ```

        ```python
        [-0.0020519258, 0.0147288125, -0.0090887165]
        ```

    """

    model_id: str | None = None
    """Type of model to use."""

    model: str | None = None
    """
    Name or alias of the foundation model to use.
    When using IBM's watsonx.ai Model Gateway (public preview), you can specify any
    supported third-party model—OpenAI, Anthropic, NVIDIA, Cerebras, or IBM's own
    Granite series—via a single, OpenAI-compatible interface. Models must be explicitly
    provisioned (opt-in) through the Gateway to ensure secure, vendor-agnostic access
    and easy switch-over without reconfiguration.

    For more details on configuration and usage, see [IBM watsonx Model Gateway docs](https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-model-gateway.html?context=wx&audience=wdp)
    """

    project_id: str | None = None
    """ID of the Watson Studio project."""

    space_id: str | None = None
    """ID of the Watson Studio space."""

    url: SecretStr = Field(
        alias="url",
        default_factory=secret_from_env("WATSONX_URL", default=None),  # type: ignore[assignment]
    )
    """URL to the Watson Machine Learning or CPD instance."""

    apikey: SecretStr | None = None
    api_key: SecretStr | None = Field(
        default_factory=secret_from_env_multi(
            names_priority=["WATSONX_API_KEY", "WATSONX_APIKEY"],
            deprecated={"WATSONX_APIKEY"},
        ),
        serialization_alias="api_key",
        validation_alias=AliasChoices("api_key", "apikey"),  # accept both on input
        description="API key to the Watson Machine Learning or CPD instance.",
    )
    """API key to the Watson Machine Learning or CPD instance."""

    token: SecretStr | None = Field(
        alias="token",
        default_factory=secret_from_env("WATSONX_TOKEN", default=None),
    )
    """Token to the CPD instance."""

    password: SecretStr | None = Field(
        alias="password",
        default_factory=secret_from_env("WATSONX_PASSWORD", default=None),
    )
    """Password to the CPD instance."""

    username: SecretStr | None = Field(
        alias="username",
        default_factory=secret_from_env("WATSONX_USERNAME", default=None),
    )
    """Username to the CPD instance."""

    instance_id: SecretStr | None = Field(
        alias="instance_id",
        default_factory=secret_from_env("WATSONX_INSTANCE_ID", default=None),
    )
    """Instance_id of the CPD instance."""

    version: SecretStr | None = None
    """Version of the CPD instance."""

    params: dict[str, Any] | None = None
    """Model parameters to use during request generation."""

    verify: str | bool | None = None
    """You can pass one of following as verify:
        * the path to a CA_BUNDLE file
        * the path of directory with certificates of trusted CAs
        * True - default path to truststore will be taken
        * False - no verification will be made
    """

    watsonx_embed: Embeddings = Field(default=None)  #: :meta private:

    watsonx_embed_gateway: Gateway = Field(
        default=None,
        exclude=True,
    )  #: :meta private:

    watsonx_client: APIClient | None = Field(default=None)  #: :meta private:

    model_config = ConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
        protected_namespaces=(),
    )

    @model_validator(mode="before")
    @classmethod
    def _normalize_and_warn_deprecated_input(cls, data: Any) -> Any:
        if isinstance(data, dict):
            # Handle deprecated input kwarg name `apikey` vs new `api_key`.
            data = normalize_api_key(data=data)
        return data

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that credentials and python package exists in environment."""
        if self.watsonx_embed_gateway is not None:
            error_msg = (
                "Passing the 'watsonx_embed_gateway' parameter to the "
                "WatsonxEmbeddings constructor is not supported yet.",
            )
            raise NotImplementedError(error_msg)

        if isinstance(self.watsonx_embed, Embeddings):
            self.model_id = self.watsonx_embed.model_id
            self.project_id = self.watsonx_embed._client.default_project_id  # noqa: SLF001
            self.space_id = self.watsonx_embed._client.default_space_id  # noqa: SLF001

            self.params = self.watsonx_embed.params

        elif isinstance(self.watsonx_client, APIClient):
            if sum(map(bool, (self.model, self.model_id))) != 1:
                error_msg = (
                    "The parameters 'model' and 'model_id' are mutually exclusive. "
                    "Please specify exactly one of these parameters when "
                    "initializing WatsonxEmbeddings.",
                )
                raise ValueError(error_msg)
            if self.model is not None:
                watsonx_embed_gateway = Gateway(
                    api_client=self.watsonx_client,
                    verify=self.verify,
                )
                self.watsonx_embed_gateway = watsonx_embed_gateway
            else:
                watsonx_embed = Embeddings(
                    model_id=self.model_id,
                    params=self.params,
                    api_client=self.watsonx_client,
                    project_id=self.project_id,
                    space_id=self.space_id,
                    verify=self.verify,
                )
                self.watsonx_embed = watsonx_embed

        else:
            if sum(map(bool, (self.model, self.model_id))) != 1:
                error_msg = (
                    "The parameters 'model' and 'model_id' are mutually exclusive. "
                    "Please specify exactly one of these parameters when "
                    "initializing WatsonxEmbeddings.",
                )
                raise ValueError(error_msg)
            credentials = resolve_watsonx_credentials(
                url=self.url,
                api_key=self.api_key,
                token=self.token,
                password=self.password,
                username=self.username,
                instance_id=self.instance_id,
                version=self.version,
                verify=self.verify,
            )
            if self.model is not None:
                watsonx_embed_gateway = Gateway(
                    credentials=credentials,
                    verify=self.verify,
                )
                self.watsonx_embed_gateway = watsonx_embed_gateway

            else:
                watsonx_embed = Embeddings(
                    model_id=self.model_id,
                    params=self.params,
                    credentials=credentials,
                    project_id=self.project_id,
                    space_id=self.space_id,
                )

                self.watsonx_embed = watsonx_embed

        return self

    @gateway_error_handler
    def _call_model_gateway(
        self,
        *,
        model: str,
        texts: list[str],
        **params: Any,
    ) -> Any:
        return self.watsonx_embed_gateway.embeddings.create(
            model=model,
            input=texts,
            **params,
        )

    @async_gateway_error_handler
    async def _acall_model_gateway(
        self,
        *,
        model: str,
        texts: list[str],
        **params: Any,
    ) -> Any:
        return await self.watsonx_embed_gateway.embeddings.acreate(
            model=model,
            input=texts,
            **params,
        )

    def embed_documents(self, texts: list[str], **kwargs: Any) -> list[list[float]]:
        """Embed search docs."""
        params = extract_params(kwargs, self.params)
        if self.watsonx_embed_gateway is not None:
            call_kwargs = {**kwargs, **params}
            embed_response = self._call_model_gateway(
                model=self.model,
                texts=texts,
                **call_kwargs,
            )
            return [embedding["embedding"] for embedding in embed_response["data"]]
        return cast(
            "list[list[float]]",
            self.watsonx_embed.embed_documents(
                texts=texts,
                **(kwargs | {"params": params}),
            ),
        )

    async def aembed_documents(
        self,
        texts: list[str],
        **kwargs: Any,
    ) -> list[list[float]]:
        """Asynchronous Embed search docs."""
        params = extract_params(kwargs, self.params)
        if self.watsonx_embed_gateway is not None:
            call_kwargs = {**kwargs, **params}
            embed_response = await self._acall_model_gateway(
                model=self.model,
                texts=texts,
                **call_kwargs,
            )
            return [embedding["embedding"] for embedding in embed_response["data"]]
        return cast(
            "list[list[float]]",
            await self.watsonx_embed.aembed_documents(
                texts=texts,
                **(kwargs | {"params": params}),
            ),
        )

    def embed_query(self, text: str, **kwargs: Any) -> list[float]:
        """Embed query text."""
        return self.embed_documents([text], **kwargs)[0]

    async def aembed_query(self, text: str, **kwargs: Any) -> list[float]:
        """Asynchronous Embed query text."""
        embeddings = await self.aembed_documents([text], **kwargs)
        return embeddings[0]



import os
from typing import Any

from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]
from langchain_core.tools import BaseTool
from langchain_tests.integration_tests.tools import ToolsIntegrationTests

from langchain_ibm.agent_toolkits.utility import WatsonxTool

WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

URL = "https://us-south.ml.cloud.ibm.com"

wx_credentials = Credentials(url=URL, api_key=WX_APIKEY)


class TestWatsonxToolsStandard(ToolsIntegrationTests):
    @property
    def tool_constructor(self) -> type[BaseTool] | BaseTool:
        return WatsonxTool

    @property
    def tool_constructor_params(self) -> dict[str, Any]:
        return {
            "name": "GoogleSearch",
            "description": "Search for online trends, news, current events, "
            "real-time information, or research topics.",
            "agent_description": "Search for online trends, news, current events, "
            "real-time information, or research topics.",
            "tool_config_schema": {
                "title": "config schema for GoogleSearch tool",
                "type": "object",
                "properties": {
                    "maxResults": {
                        "title": "Max number of results to return",
                        "type": "integer",
                        "minimum": 1,
                        "maximum": 20,
                    }
                },
            },
            "tool_input_schema": {
                "type": "object",
                "properties": {
                    "q": {
                        "title": "Query",
                        "description": "GoogleSearch query",
                        "type": "string",
                    }
                },
                "required": ["q"],
            },
            "watsonx_client": APIClient(
                credentials=wx_credentials,
                project_id=WX_PROJECT_ID,
            ),
        }

    @property
    def tool_invoke_params_example(self) -> dict[str, Any]:
        """
        Returns a dictionary representing the "args" of an example tool call.

        This should NOT be a ToolCall dict - i.e. it should not
        have {"name", "id", "args"} keys.
        """
        return {
            "q": "Search IBM",
        }



"""Test ChatWatsonx API wrapper."""

import json
import os
import re
import sys
from typing import Any
from unittest.mock import Mock

import pytest
from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models import (  # type: ignore[import-untyped]
    ModelInference,
)
from ibm_watsonx_ai.foundation_models.schema import (  # type: ignore[import-untyped]
    TextChatParameters,
)
from ibm_watsonx_ai.gateway import Gateway  # type: ignore[import-untyped]
from ibm_watsonx_ai.wml_client_error import (  # type: ignore[import-untyped]
    WMLClientError,
)

from langchain_ibm import ChatWatsonx
from langchain_ibm.chat_models import normalize_tool_arguments

os.environ.pop("WATSONX_APIKEY", None)
os.environ.pop("WATSONX_PROJECT_ID", None)

MODEL_ID = "mistralai/mixtral-8x7b-instruct-v01"

api_client_mock = Mock(spec=APIClient)
api_client_mock.default_space_id = None
api_client_mock.default_project_id = "fake_project_id"

gateway_mock = Mock(spec=Gateway)
gateway_mock._client = api_client_mock

model_inference_mock = Mock(spec=ModelInference)
model_inference_mock._client = api_client_mock
model_inference_mock.model_id = "fake_model_id"
model_inference_mock.params = {"temperature": 1}


def test_initialize_chat_watsonx_without_url() -> None:
    pattern = r"(?=.*url)(?=.*WATSONX_URL)"
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
        )


def test_initialize_chat_watsonx_cloud_only_url() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(model_id=MODEL_ID, url="https://us-south.ml.cloud.ibm.com")


def test_initialize_chat_watsonx_with_deprecated_apikey() -> None:
    with (
        pytest.warns(
            DeprecationWarning,
            match="'apikey' parameter is deprecated; use 'api_key' instead.",
        ),
        pytest.raises(WMLClientError),
    ):
        ChatWatsonx(
            model_id=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_chat_watsonx_with_api_key_and_apikey() -> None:
    with (
        pytest.warns(
            UserWarning,
            match="Both 'api_key' and deprecated 'apikey' were provided; "
            "'api_key' takes precedence.",
        ),
        pytest.raises(WMLClientError),
    ):
        ChatWatsonx(
            model_id=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="fake_apikey",
            api_key="fake_api_key",
        )


def test_initialize_chat_watsonx_cpd_without_all() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*password)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_PASSWORD)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
        )


def test_initialize_chat_watsonx_cpd_only_password() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            password="fake_password",  # noqa: S106
        )


def test_initialize_chat_watsonx_cpd_only_username() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*password)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_PASSWORD)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            username="fake_username",
        )


def test_initialize_chat_watsonx_cpd_only_paikey() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="fake_apikey",
        )


def test_initialize_chat_watsonx_cpd_deprecation_warning_with_instance_id() -> None:
    with (
        pytest.warns(
            DeprecationWarning, match="The `instance_id` parameter is deprecated"
        ),
        pytest.raises(WMLClientError),
    ):
        ChatWatsonx(
            model_id="google/flan-ul2",
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
            username="test_user",
            instance_id="openshift",
        )


def test_initialize_chat_watsonx_with_two_exclusive_parameters() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing ChatWatsonx."
    )
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
            model=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_chat_watsonx_with_three_exclusive_parameters() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing ChatWatsonx."
    )
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(
            model_id=MODEL_ID,
            model=MODEL_ID,
            deployment_id="test_deployment_id",
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_chat_watsonx_with_api_client_only() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing ChatWatsonx."
    )
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx(watsonx_client=api_client_mock)


def test_initialize_chat_watsonx_with_watsonx_model_gateway() -> None:
    pattern = re.escape(
        "Passing the 'watsonx_model_gateway' parameter to the ChatWatsonx "
        "constructor is not supported yet."
    )
    with pytest.raises(NotImplementedError, match=pattern):
        ChatWatsonx(watsonx_model_gateway=gateway_mock)


def test_initialize_chat_watsonx_without_any_params() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing ChatWatsonx."
    )
    with pytest.raises(ValueError, match=pattern):
        ChatWatsonx()


def test_initialize_chat_watsonx_with_model_inference_only() -> None:
    chat = ChatWatsonx(watsonx_model=model_inference_mock)

    assert isinstance(chat, ChatWatsonx)


@pytest.mark.skipif(
    sys.version_info[:2] != (3, 10),
    reason="Test runs only on Python 3.10",
)
def test_initialize_chat_watsonx_with_all_supported_params(mocker: Any) -> None:
    top_p = 0.8

    def mock_modelinference_chat(**kwargs: Any) -> dict[str, Any]:
        assert kwargs.get("messages") == [{"content": "Hello", "role": "user"}]
        assert kwargs.get("params") == (
            {
                k: v
                for k, v in TextChatParameters.get_sample_params().items()
                if "guided" not in k
                if "chat_template_kwargs" not in k
                if "reasoning_effort" not in k
                if "include_reasoning" not in k
                if "length_penalty" not in k
                if "repetition_penalty" not in k
            }
            | {
                "logit_bias": {"1003": -100, "1004": -100},
                "seed": 41,
                "stop": ["this", "the"],
            }
            | {"top_p": top_p}
        )

        return {"id": "123", "choices": [{"message": {"content": "Hi", "role": "ai"}}]}

    mocker.patch(
        "ibm_watsonx_ai.foundation_models.ModelInference.__init__",
        return_value=None,
    )
    mocker.patch(
        "ibm_watsonx_ai.foundation_models.ModelInference.chat",
        side_effect=mock_modelinference_chat,
    )

    chat = ChatWatsonx(
        model_id="google/flan-ul2",
        url="https://us-south.ml.cloud.ibm.com",
        apikey="test_apikey",
        frequency_penalty=0.5,
        logprobs=True,
        top_logprobs=3,
        presence_penalty=0.3,
        response_format={
            "type": "json_schema",
            "json_schema": {
                "name": "Sample JSON schema",
                "schema": {
                    "title": "SimpleUser",
                    "type": "object",
                    "properties": {
                        "username": {"type": "string"},
                        "email": {"type": "string", "format": "email"},
                    },
                    "required": ["username", "email"],
                },
                "strict": False,
            },
        },
        temperature=0.7,
        max_completion_tokens=512,
        time_limit=600000,
        top_p=0.9,
        n=1,
        logit_bias={"1003": -100, "1004": -100},
        seed=41,
        stop=["this", "the"],
    )

    # change only top_n
    chat.invoke("Hello", top_p=top_p)


@pytest.mark.skipif(
    not ((3, 11) <= sys.version_info[:2] <= (3, 13)),
    reason="Test runs only on Python 3.11-3.13",
)
def test_initialize_chat_watsonx_with_all_supported_params_v2(mocker: Any) -> None:
    top_p = 0.8

    def mock_modelinference_chat(**kwargs: Any) -> dict[str, Any]:
        assert kwargs.get("messages") == [{"content": "Hello", "role": "user"}]
        assert kwargs.get("params") == (
            {
                k: v
                for k, v in TextChatParameters.get_sample_params().items()
                if "guided" not in k
            }
            | {
                "logit_bias": {"1003": -100, "1004": -100},
                "seed": 41,
                "stop": ["this", "the"],
            }
            | {"top_p": top_p}
        )

        return {"id": "123", "choices": [{"message": {"content": "Hi", "role": "ai"}}]}

    mocker.patch(
        "ibm_watsonx_ai.foundation_models.ModelInference.__init__",
        return_value=None,
    )
    mocker.patch(
        "ibm_watsonx_ai.foundation_models.ModelInference.chat",
        side_effect=mock_modelinference_chat,
    )

    chat = ChatWatsonx(
        model_id="google/flan-ul2",
        url="https://us-south.ml.cloud.ibm.com",
        apikey="test_apikey",
        frequency_penalty=0.5,
        logprobs=True,
        top_logprobs=3,
        presence_penalty=0.3,
        response_format={
            "type": "json_schema",
            "json_schema": {
                "name": "Sample JSON schema",
                "schema": {
                    "title": "SimpleUser",
                    "type": "object",
                    "properties": {
                        "username": {"type": "string"},
                        "email": {"type": "string", "format": "email"},
                    },
                    "required": ["username", "email"],
                },
                "strict": False,
            },
        },
        temperature=0.7,
        max_completion_tokens=512,
        time_limit=600000,
        top_p=0.9,
        n=1,
        logit_bias={"1003": -100, "1004": -100},
        seed=41,
        stop=["this", "the"],
        length_penalty=1.0,
        repetition_penalty=1.5,
        reasoning_effort="high",
        include_reasoning=True,
        chat_template_kwargs={"thinking": True},
    )

    # change only top_n
    chat.invoke("Hello", top_p=top_p)


# Tests for normalize_tool_arguments function
def test_normalize_tool_arguments_with_json_string() -> None:
    """Test normalizing JSON string arguments."""
    # Test case 1: JSON string
    json_str = '{"location": "San Francisco", "unit": "celsius"}'
    result = normalize_tool_arguments(json_str)
    assert result == '{"location": "San Francisco", "unit": "celsius"}'


def test_normalize_tool_arguments_with_python_dict_string() -> None:
    """Test normalizing Python dict string arguments."""
    # Test case 2: Python dict string with single quotes
    python_dict_str = "{'location': 'San Francisco', 'unit': 'celsius'}"
    result = normalize_tool_arguments(python_dict_str)
    assert result == '{"location": "San Francisco", "unit": "celsius"}'


def test_normalize_tool_arguments_with_extra_quotes() -> None:
    """Test normalizing arguments with extra surrounding quotes."""
    # Test case 3: Extra wrapping quotes like '"{...}"'
    wrapped_json = '"{\\"location\\": \\"San Francisco\\", \\"unit\\": \\"celsius\\"}"'
    result = normalize_tool_arguments(wrapped_json)
    assert result == '{"location": "San Francisco", "unit": "celsius"}'


def test_normalize_tool_arguments_with_nested_structures() -> None:
    """Test normalizing arguments with nested dict/list structures."""
    # Test case 4: Nested structures
    nested_str = '{"user": {"name": "John", "prefs": ["temp", "humidity"]}}'
    result = normalize_tool_arguments(nested_str)
    assert result == '{"user": {"name": "John", "prefs": ["temp", "humidity"]}}'


def test_normalize_tool_arguments_with_empty_dict() -> None:
    """Test normalizing empty dict arguments."""
    # Test case 5: Empty dict
    empty_dict = "{}"
    result = normalize_tool_arguments(empty_dict)
    assert result == "{}"


def test_normalize_tool_arguments_with_already_valid_json() -> None:
    """Test that already valid JSON is returned as-is."""
    # Test case 6: Already valid JSON
    valid_json = '{"key": "value", "number": 42, "bool": true}'
    result = normalize_tool_arguments(valid_json)
    assert result == '{"key": "value", "number": 42, "bool": true}'


def test_normalize_tool_arguments_with_special_characters() -> None:
    """Test normalizing arguments with special characters."""
    # Test case 7: Special characters and escaped strings
    special_chars = '{"message": "Hello \\"world\\"!", "path": "C:\\\\\\\\Users"}'
    result = normalize_tool_arguments(special_chars)

    parsed = json.loads(result)
    assert isinstance(parsed, dict)
    assert "message" in parsed


def test_normalize_tool_arguments_with_numbers_and_booleans() -> None:
    """Test normalizing arguments with various data types."""
    # Test case 8: Mixed data types
    mixed_types = '{"temp": 25.5, "enabled": true, "count": 10, "data": null}'
    result = normalize_tool_arguments(mixed_types)

    parsed = json.loads(result)
    assert parsed["temp"] == 25.5
    assert parsed["enabled"] is True
    assert parsed["count"] == 10
    assert parsed["data"] is None



"""IBM watsonx.ai chat wrapper."""

from __future__ import annotations

import ast
import contextlib
import json
import logging
import warnings
from operator import itemgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    cast,
)

from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models import (  # type: ignore[import-untyped]
    ModelInference,
)
from ibm_watsonx_ai.foundation_models.schema import (  # type: ignore[import-untyped]
    BaseSchema,
    TextChatParameters,
)
from ibm_watsonx_ai.gateway import Gateway  # type: ignore[import-untyped]
from langchain_core.language_models.chat_models import (  # type: ignore[attr-defined]
    BaseChatModel,
    LangSmithParams,
    agenerate_from_stream,
    generate_from_stream,
)
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    ChatMessage,
    ChatMessageChunk,
    FunctionMessage,
    FunctionMessageChunk,
    HumanMessage,
    HumanMessageChunk,
    InvalidToolCall,
    SystemMessage,
    SystemMessageChunk,
    ToolCall,
    ToolMessage,
    ToolMessageChunk,
)
from langchain_core.messages.ai import UsageMetadata
from langchain_core.messages.tool import tool_call_chunk
from langchain_core.output_parsers import JsonOutputParser, PydanticOutputParser
from langchain_core.output_parsers.openai_tools import (
    JsonOutputKeyToolsParser,
    PydanticToolsParser,
    make_invalid_tool_call,
    parse_tool_call,
)
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from langchain_core.runnables import (
    Runnable,
    RunnableMap,
    RunnablePassthrough,
)
from langchain_core.utils._merge import merge_dicts
from langchain_core.utils.function_calling import (
    convert_to_openai_function,
    convert_to_openai_tool,
)
from langchain_core.utils.pydantic import (
    TypeBaseModel,
    is_basemodel_subclass,
)
from langchain_core.utils.utils import secret_from_env
from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    SecretStr,
    model_validator,
)
from pydantic.v1 import BaseModel as BaseModelV1
from typing_extensions import Self, override

from langchain_ibm.utils import (
    async_gateway_error_handler,
    check_duplicate_chat_params,
    extract_chat_params,
    gateway_error_handler,
    normalize_api_key,
    resolve_watsonx_credentials,
    secret_from_env_multi,
)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable, Iterator, Mapping, Sequence

    from langchain_core.callbacks import (
        AsyncCallbackManagerForLLMRun,
        CallbackManagerForLLMRun,
    )
    from langchain_core.language_models import LanguageModelInput
    from langchain_core.tools import BaseTool

logger = logging.getLogger(__name__)


def normalize_tool_arguments(args_str: str) -> str:
    """Ensure arguments is always a proper JSON string.

    Handles:
    - JSON string
    - Python dict string
    - Extra wrapping quotes like '"{...}"'
    Args:
        args_str: tool call args_str.

    Returns:
        The LangChain tool call arguments args_str.
    """
    # Try to parse as JSON
    try:
        parsed = json.loads(args_str)
    except json.JSONDecodeError:
        pass
    else:
        if isinstance(parsed, str):
            json.loads(parsed)
            return parsed
        return args_str

    # Try Python literal (e.g., "{'a': 1}")
    obj: Any = ast.literal_eval(args_str)
    return json.dumps(obj, ensure_ascii=False)


def _convert_dict_to_message(_dict: Mapping[str, Any], call_id: str) -> BaseMessage:
    """Convert a dictionary to a LangChain message.

    Args:
        _dict: The dictionary.
        call_id: call id

    Returns:
        The LangChain message.
    """
    role = _dict.get("role")
    name = _dict.get("name")
    id_ = call_id
    if role == "user":
        return HumanMessage(content=_dict.get("content", ""), id=id_, name=name)
    if role == "assistant":
        content = _dict.get("content", "") or ""
        additional_kwargs: dict[str, Any] = {}
        if function_call := _dict.get("function_call"):
            additional_kwargs["function_call"] = dict(function_call)
        tool_calls = []
        invalid_tool_calls = []
        if raw_tool_calls := _dict.get("tool_calls"):
            for raw_tool_call in raw_tool_calls:
                ## Code change to support langgraph with A2A and graph.astream.
                if "function" in raw_tool_call:
                    func = raw_tool_call.get("function", {})
                    if "arguments" in func:
                        raw_args = raw_tool_call["function"]["arguments"]
                        json_args_str = normalize_tool_arguments(raw_args)
                        raw_tool_call["function"]["arguments"] = json_args_str

                try:
                    tool_calls.append(parse_tool_call(raw_tool_call, return_id=True))
                except Exception as e:
                    invalid_tool_calls.append(
                        make_invalid_tool_call(raw_tool_call, str(e)),
                    )
        if audio := _dict.get("audio"):
            additional_kwargs["audio"] = audio
        if reasoning_content := _dict.get("reasoning_content"):
            additional_kwargs["reasoning_content"] = reasoning_content
        return AIMessage(
            content=content,
            additional_kwargs=additional_kwargs,
            name=name,
            id=id_,
            tool_calls=tool_calls,
            invalid_tool_calls=invalid_tool_calls,
        )
    if role == "system":
        return SystemMessage(content=_dict.get("content", ""), name=name, id=id_)
    if role == "function":
        return FunctionMessage(
            content=_dict.get("content", ""),
            name=cast("str", _dict.get("name")),
            id=id_,
        )
    if role == "tool":
        additional_kwargs = {}
        if "name" in _dict:
            additional_kwargs["name"] = _dict["name"]
        return ToolMessage(
            content=_dict.get("content", ""),
            tool_call_id=cast("str", _dict.get("tool_call_id")),
            additional_kwargs=additional_kwargs,
            name=name,
            id=id_,
        )
    return ChatMessage(content=_dict.get("content", ""), role=role, id=id_)  # type: ignore[unused-ignore]


def _format_message_content(content: Any) -> Any:
    """Format message content."""
    if content and isinstance(content, list):
        formatted_content = []
        for block in content:
            # Remove unexpected block types
            if (
                isinstance(block, dict)
                and "type" in block
                and block["type"] in {"tool_use", "thinking", "reasoning_content"}
            ):
                continue

            # Image blocks
            if isinstance(block, dict) and block.get("type") == "image":
                if (data := block.get("base64")) and (
                    mime_type := block.get("mime_type")
                ):
                    formatted_content.append(
                        {
                            "type": "image_url",
                            "image_url": {"url": f"data:{mime_type};base64,{data}"},
                        }
                    )
                else:
                    continue
            else:
                formatted_content.append(block)
    else:
        formatted_content = content

    return formatted_content


def _convert_message_to_dict(message: BaseMessage) -> dict[str, Any]:
    """Convert a LangChain message to a dictionary.

    Args:
        message: The LangChain message.

    Returns:
        The dictionary.
    """
    message_dict: dict[str, Any] = {"content": _format_message_content(message.content)}
    if (name := message.name or message.additional_kwargs.get("name")) is not None:
        message_dict["name"] = name

    # populate role and additional message data
    if isinstance(message, ChatMessage):
        message_dict["role"] = message.role
    elif isinstance(message, HumanMessage):
        message_dict["role"] = "user"
    elif isinstance(message, AIMessage):
        message_dict["role"] = "assistant"
        if message.tool_calls or message.invalid_tool_calls:
            message_dict["tool_calls"] = [
                _lc_tool_call_to_watsonx_tool_call(tc) for tc in message.tool_calls
            ] + [
                _lc_invalid_tool_call_to_watsonx_tool_call(tc)
                for tc in message.invalid_tool_calls
            ]
        elif "tool_calls" in message.additional_kwargs:
            message_dict["tool_calls"] = message.additional_kwargs["tool_calls"]
            tool_call_supported_props = {"id", "type", "function"}
            message_dict["tool_calls"] = [
                {k: v for k, v in tool_call.items() if k in tool_call_supported_props}
                for tool_call in message_dict["tool_calls"]
            ]
        elif "function_call" in message.additional_kwargs:
            message_dict["function_call"] = message.additional_kwargs["function_call"]
        else:
            pass
        # If tool calls present, content null value should be None not empty string.
        if "function_call" in message_dict or "tool_calls" in message_dict:
            message_dict["content"] = message_dict["content"] or None

        audio: dict[str, Any] | None = None
        for block in message.content:
            if (
                isinstance(block, dict)
                and block.get("type") == "audio"
                and (id_ := block.get("id"))
            ):
                audio = {"id": id_}
        if not audio and "audio" in message.additional_kwargs:
            raw_audio = message.additional_kwargs["audio"]
            audio = (
                {"id": message.additional_kwargs["audio"]["id"]}
                if "id" in raw_audio
                else raw_audio
            )
        if audio:
            message_dict["audio"] = audio
    elif isinstance(message, SystemMessage):
        message_dict["role"] = "system"
    elif isinstance(message, FunctionMessage):
        message_dict["role"] = "function"
    elif isinstance(message, ToolMessage):
        message_dict["role"] = "tool"
        message_dict["tool_call_id"] = message.tool_call_id

        supported_props = {"content", "role", "tool_call_id"}
        message_dict = {k: v for k, v in message_dict.items() if k in supported_props}
    else:
        error_msg = f"Got unknown type {message}"
        raise TypeError(error_msg)
    return message_dict


def _convert_delta_to_message_chunk(
    _dict: Mapping[str, Any],
    default_class: type[BaseMessageChunk],
    call_id: str,
    *,
    is_first_tool_chunk: bool,
) -> BaseMessageChunk:
    id_ = call_id
    role = cast("str", _dict.get("role"))
    content = cast("str", _dict.get("content") or "")
    additional_kwargs: dict[str, Any] = {}
    if _dict.get("function_call"):
        function_call = dict(_dict["function_call"])
        if "name" in function_call and function_call["name"] is None:
            function_call["name"] = ""
        additional_kwargs["function_call"] = function_call
    tool_call_chunks = []
    if raw_tool_calls := _dict.get("tool_calls"):
        additional_kwargs["tool_calls"] = raw_tool_calls
        with contextlib.suppress(KeyError):
            tool_call_chunks = [
                tool_call_chunk(
                    name=rtc["function"].get("name")
                    if is_first_tool_chunk or (rtc.get("id") is not None)
                    else None,
                    args=rtc["function"].get("arguments"),
                    # `id` is provided only for the first delta with unique tool_calls
                    # (multiple tool calls scenario)
                    id=rtc.get("id"),
                    index=rtc["index"],
                )
                for rtc in raw_tool_calls
            ]

    if reasoning_content := _dict.get("reasoning_content"):
        additional_kwargs["reasoning_content"] = reasoning_content

    if role == "user" or default_class == HumanMessageChunk:
        return HumanMessageChunk(content=content, id=id_)
    if role == "assistant" or default_class == AIMessageChunk:
        return AIMessageChunk(
            content=content,
            additional_kwargs=additional_kwargs,
            id=id_,
            tool_call_chunks=tool_call_chunks,  # type: ignore[unused-ignore]
        )
    if role == "system" or default_class == SystemMessageChunk:
        return SystemMessageChunk(content=content, id=id_)
    if role == "function" or default_class == FunctionMessageChunk:
        return FunctionMessageChunk(content=content, name=_dict["name"], id=id_)
    if role == "tool" or default_class == ToolMessageChunk:
        return ToolMessageChunk(
            content=content,
            tool_call_id=_dict["tool_call_id"],
            id=id_,
        )
    if role or default_class == ChatMessageChunk:
        return ChatMessageChunk(content=content, role=role, id=id_)
    return default_class(content=content, id=id_)  # type: ignore[call-arg]


def _convert_chunk_to_generation_chunk(
    chunk: dict[str, Any],
    default_chunk_class: type,
    *,
    is_first_tool_chunk: bool,
    _prompt_tokens_included: bool,
) -> ChatGenerationChunk | None:
    token_usage = chunk.get("usage")
    choices = chunk.get("choices", [])

    usage_metadata: UsageMetadata | None = (
        _create_usage_metadata(
            token_usage, _prompt_tokens_included=_prompt_tokens_included
        )
        if token_usage
        else None
    )

    if len(choices) == 0:
        # logprobs is implicitly None
        return ChatGenerationChunk(
            message=default_chunk_class(content="", usage_metadata=usage_metadata),
        )

    choice = choices[0]
    if choice["delta"] is None:
        return None

    message_chunk = _convert_delta_to_message_chunk(
        choice["delta"],
        default_chunk_class,
        chunk["id"],
        is_first_tool_chunk=is_first_tool_chunk,
    )
    generation_info = {}

    if finish_reason := choice.get("finish_reason"):
        generation_info["finish_reason"] = finish_reason
        if model_name := chunk.get("model"):
            generation_info["model_name"] = model_name
        if system_fingerprint := chunk.get("system_fingerprint"):
            generation_info["system_fingerprint"] = system_fingerprint

    logprobs = choice.get("logprobs")
    if logprobs:
        generation_info["logprobs"] = logprobs

    if usage_metadata and isinstance(message_chunk, AIMessageChunk):
        message_chunk.usage_metadata = usage_metadata

    return ChatGenerationChunk(
        message=message_chunk,
        generation_info=generation_info or None,
    )


class ChatWatsonx(BaseChatModel):
    r"""`IBM watsonx.ai` chat models integration.

    ???+ info "Setup"

        To use, you should have `langchain_ibm` python package installed,
        and the environment variable `WATSONX_API_KEY` set with your API key, or pass
        it as a named parameter `api_key` to the constructor.

        ```bash
        pip install -U langchain-ibm

        # or using uv
        uv add langchain-ibm
        ```

        ```bash
        export WATSONX_API_KEY="your-api-key"
        ```

        !!! deprecated
            `apikey` and `WATSONX_APIKEY` are deprecated and will be removed in
            version `2.0.0`. Use `api_key` and `WATSONX_API_KEY` instead.

    ??? info "Instantiate"

        Create a model instance with desired params. For example:

        ```python
        from langchain_ibm import ChatWatsonx
        from ibm_watsonx_ai.foundation_models.schema import TextChatParameters

        parameters = TextChatParameters(
            top_p=1, temperature=0.5, max_completion_tokens=None
        )

        model = ChatWatsonx(
            model_id="meta-llama/llama-3-3-70b-instruct",
            url="https://us-south.ml.cloud.ibm.com",
            project_id="*****",
            params=parameters,
            # api_key="*****"
        )
        ```

    ??? info "Invoke"

        Generate a response from the model:

        ```python
        messages = [
            (
                "system",
                "You are a helpful translator. Translate the user sentence to French.",
            ),
            ("human", "I love programming."),
        ]
        model.invoke(messages)
        ```

        Results in an `AIMessage` response:

        ```python
        AIMessage(
            content="J'adore programmer.",
            additional_kwargs={},
            response_metadata={
                "token_usage": {
                    "completion_tokens": 7,
                    "prompt_tokens": 30,
                    "total_tokens": 37,
                },
                "model_name": "ibm/granite-3-3-8b-instruct",
                "system_fingerprint": "",
                "finish_reason": "stop",
            },
            id="chatcmpl-529352c4-93ba-4801-8f1d-a3b4e3935194---daed91fb74d0405f200db1e63da9a48a---7a3ef799-4413-47e4-b24c-85d267e37fa2",
            usage_metadata={"input_tokens": 30, "output_tokens": 7, "total_tokens": 37},
        )
        ```

    ??? info "Stream"

        Stream a response from the model:

        ```python
        for chunk in model.stream(messages):
            print(chunk.text)
        ```

        Results in a sequence of `AIMessageChunk` objects with partial content:

        ```python
        AIMessageChunk(content="", id="run--e48a38d3-1500-4b5e-870c-6313e8cff775")
        AIMessageChunk(content="J", id="run--e48a38d3-1500-4b5e-870c-6313e8cff775")
        AIMessageChunk(content="'", id="run--e48a38d3-1500-4b5e-870c-6313e8cff775")
        AIMessageChunk(content="ad", id="run--e48a38d3-1500-4b5e-870c-6313e8cff775")
        AIMessageChunk(content="or", id="run--e48a38d3-1500-4b5e-870c-6313e8cff775")
        AIMessageChunk(
            content=" programmer", id="run--e48a38d3-1500-4b5e-870c-6313e8cff775"
        )
        AIMessageChunk(content=".", id="run--e48a38d3-1500-4b5e-870c-6313e8cff775")
        AIMessageChunk(
            content="",
            response_metadata={
                "finish_reason": "stop",
                "model_name": "ibm/granite-3-3-8b-instruct",
            },
            id="run--e48a38d3-1500-4b5e-870c-6313e8cff775",
        )
        AIMessageChunk(
            content="",
            id="run--e48a38d3-1500-4b5e-870c-6313e8cff775",
            usage_metadata={"input_tokens": 30, "output_tokens": 7, "total_tokens": 37},
        )
        ```

        To collect the full message, you can concatenate the chunks:

        ```python
        stream = model.stream(messages)
        full = next(stream)
        for chunk in stream:
            full += chunk

        full
        ```

        ```python
        AIMessageChunk(
            content="J'adore programmer.",
            response_metadata={
                "finish_reason": "stop",
                "model_name": "ibm/granite-3-3-8b-instruct",
            },
            id="chatcmpl-88a48b71-c149-4a0c-9c02-d6b97ca5dc6c---b7ba15879a8c5283b1e8a3b8db0229f0---0037ca4f-8a74-4f84-a46c-ab3fd1294f24",
            usage_metadata={"input_tokens": 30, "output_tokens": 7, "total_tokens": 37},
        )
        ```

    ??? info "Async"

        Asynchronous equivalents of `invoke`, `stream`, and `batch` are also available:

        ```python
        # Invoke
        await model.ainvoke(messages)

        # Stream
        async for chunk in model.astream(messages):
            print(chunk.text)

        # Batch
        await model.abatch([messages])
        ```

        Results in an `AIMessage` response:

        ```python
        AIMessage(
            content="J'adore programmer.",
            additional_kwargs={},
            response_metadata={
                "token_usage": {
                    "completion_tokens": 7,
                    "prompt_tokens": 30,
                    "total_tokens": 37,
                },
                "model_name": "ibm/granite-3-3-8b-instruct",
                "system_fingerprint": "",
                "finish_reason": "stop",
            },
            id="chatcmpl-5bef2d81-ef56-463b-a8fa-c2bcc2a3c348---821e7750d18925f2b36226db66667e26---6396c786-9da9-4468-883e-11ed90a05937",
            usage_metadata={"input_tokens": 30, "output_tokens": 7, "total_tokens": 37},
        )
        ```

        For batched calls, results in a `list[AIMessage]`.

    ??? info "Tool calling"

        ```python
        from pydantic import BaseModel, Field


        class GetWeather(BaseModel):
            '''Get the current weather in a given location'''

            location: str = Field(
                ..., description="The city and state, e.g. San Francisco, CA"
            )


        class GetPopulation(BaseModel):
            '''Get the current population in a given location'''

            location: str = Field(
                ..., description="The city and state, e.g. San Francisco, CA"
            )


        model_with_tools = model.bind_tools(
            [GetWeather, GetPopulation]
            # strict = True  # Enforce tool args schema is respected
        )
        ai_msg = model_with_tools.invoke(
            "Which city is hotter today and which is bigger: LA or NY?"
        )
        ai_msg.tool_calls
        ```

        ```python
        [
            {
                "name": "GetWeather",
                "args": {"location": "Los Angeles, CA"},
                "id": "chatcmpl-tool-59632abcee8f48a18a5f3a81422b917b",
                "type": "tool_call",
            },
            {
                "name": "GetWeather",
                "args": {"location": "New York, NY"},
                "id": "chatcmpl-tool-c6f3b033b4594918bb53f656525b0979",
                "type": "tool_call",
            },
            {
                "name": "GetPopulation",
                "args": {"location": "Los Angeles, CA"},
                "id": "chatcmpl-tool-175a23281e4747ea81cbe472b8e47012",
                "type": "tool_call",
            },
            {
                "name": "GetPopulation",
                "args": {"location": "New York, NY"},
                "id": "chatcmpl-tool-e1ccc534835945aebab708eb5e685bf7",
                "type": "tool_call",
            },
        ]
        ```

    ??? info "Reasoning output"

        ```python
        from langchain_ibm import ChatWatsonx
        from ibm_watsonx_ai.foundation_models.schema import TextChatParameters

        parameters = TextChatParameters(
            include_reasoning=True, reasoning_effort="medium"
        )

        model = ChatWatsonx(
            model_id="openai/gpt-oss-120b",
            url="https://us-south.ml.cloud.ibm.com",
            project_id="*****",
            params=parameters,
            # api_key="*****"
        )

        response = model.invoke("What is 3^3?")

        # Response text
        print(f"Output: {response.content}")

        # Reasoning summaries
        print(f"Reasoning: {response.additional_kwargs['reasoning_content']}")
        ```

        ```txt
        Output: 3^3 = 27
        Reasoning: The user asks "What is 3^3?" That's 27. Provide answer.
        ```

        !!! version-added "Added in version 0.3.19: Updated `AIMessage` format"
            [`langchain-ibm >= 0.3.19`](https://pypi.org/project/langchain-ibm/#history)
            allows users to set Reasoning output parameters and will format output from
            reasoning summaries into `additional_kwargs` field.

    ??? info "Structured output"

        ```python
        from pydantic import BaseModel, Field


        class Joke(BaseModel):
            '''Joke to tell user.'''

            setup: str = Field(description="The setup of the joke")
            punchline: str = Field(description="The punchline to the joke")
            rating: int | None = Field(description="How funny the joke is, 1 to 10")


        structured_model = model.with_structured_output(Joke)
        structured_model.invoke("Tell me a joke about cats")
        ```

        ```python
        Joke(
            setup="Why was the cat sitting on the computer?",
            punchline="To keep an eye on the mouse!",
            rating=None,
        )
        ```

        See `with_structured_output` for more info.

    ??? info "JSON mode"

        ```python
        json_model = model.bind(response_format={"type": "json_object"})
        ai_msg = json_model.invoke(
            “Return JSON with 'random_ints': an array of 10 random integers from 0-99.”
        )
        ai_msg.content
        ```

        ```txt
        '{\n  "random_ints": [12, 34, 56, 78, 10, 22, 44, 66, 88, 99]\n}'
        ```

    ??? info "Image input"

        ```python
        import base64
        import httpx
        from langchain.messages import HumanMessage

        image_url = "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
        image_data = base64.b64encode(httpx.get(image_url).content).decode("utf-8")
        message = HumanMessage(
            content=[
                {"type": "text", "text": "describe the weather in this image"},
                {
                    "type": "image_url",
                    "image_url": {"url": f"data:image/jpeg;base64,{image_data}"},
                },
            ]
        )

        ai_msg = model.invoke([message])
        ai_msg.content
        ```

        ```txt
        "The weather in the image presents a clear, sunny day with good visibility
        and no immediate signs of rain or strong winds. The vibrant blue sky with
        scattered white clouds gives the impression of a tranquil, pleasant day
        conducive to outdoor activities."
        ```

    ??? info "Token usage"

        ```python
        ai_msg = model.invoke(messages)
        ai_msg.usage_metadata
        ```

        ```txt
        {'input_tokens': 30, 'output_tokens': 7, 'total_tokens': 37}
        ```

        ```python
        stream = model.stream(messages)
        full = next(stream)
        for chunk in stream:
            full += chunk
        full.usage_metadata
        ```

        ```txt
        {'input_tokens': 30, 'output_tokens': 7, 'total_tokens': 37}
        ```

    ??? info "Logprobs"

        ```python
        logprobs_model = model.bind(logprobs=True)
        ai_msg = logprobs_model.invoke(messages)
        ai_msg.response_metadata["logprobs"]
        ```

        ```txt
        {
            'content': [
                {
                    'token': 'J',
                    'logprob': -0.0017940393
                },
                {
                    'token': "'",
                    'logprob': -1.7523613e-05
                },
                {
                    'token': 'ad',
                    'logprob': -0.16112353
                },
                {
                    'token': 'ore',
                    'logprob': -0.0003091811
                },
                {
                    'token': ' programmer',
                    'logprob': -0.24849245
                },
                {
                    'token': '.',
                    'logprob': -2.5033638e-05
                },
                {
                    'token': '<|end_of_text|>',
                    'logprob': -7.080781e-05
                }
            ]
        }
        ```

    ??? info "Response metadata"

        ```python
        ai_msg = model.invoke(messages)
        ai_msg.response_metadata
        ```

        ```txt
        {
            'token_usage': {
                'completion_tokens': 7,
                'prompt_tokens': 30,
                'total_tokens': 37
            },
            'model_name': 'ibm/granite-3-3-8b-instruct',
            'system_fingerprint': '',
            'finish_reason': 'stop'
        }
        ```
    """

    model_id: str | None = None
    """Type of model to use."""

    model: str | None = None
    """
    Name or alias of the foundation model to use.
    When using IBM's watsonx.ai Model Gateway (public preview), you can specify any
    supported third-party model—OpenAI, Anthropic, NVIDIA, Cerebras, or IBM's own
    Granite series—via a single, OpenAI-compatible interface. Models must be explicitly
    provisioned (opt-in) through the Gateway to ensure secure, vendor-agnostic access
    and easy switch-over without reconfiguration.

    For more details on configuration and usage, see [IBM watsonx Model Gateway docs](https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-model-gateway.html?context=wx&audience=wdp)
    """

    deployment_id: str | None = None
    """Type of deployed model to use."""

    project_id: str | None = None
    """ID of the Watson Studio project."""

    space_id: str | None = None
    """ID of the Watson Studio space."""

    url: SecretStr = Field(
        alias="url",
        default_factory=secret_from_env("WATSONX_URL", default=None),  # type: ignore[assignment]
    )
    """URL to the Watson Machine Learning or CPD instance."""

    apikey: SecretStr | None = None
    api_key: SecretStr | None = Field(
        default_factory=secret_from_env_multi(
            names_priority=["WATSONX_API_KEY", "WATSONX_APIKEY"],
            deprecated={"WATSONX_APIKEY"},
        ),
        serialization_alias="api_key",
        validation_alias=AliasChoices("api_key", "apikey"),  # accept both on input
        description="API key to the Watson Machine Learning or CPD instance.",
    )
    """API key to the Watson Machine Learning or CPD instance."""

    token: SecretStr | None = Field(
        alias="token",
        default_factory=secret_from_env("WATSONX_TOKEN", default=None),
    )
    """Token to the CPD instance."""

    password: SecretStr | None = Field(
        alias="password",
        default_factory=secret_from_env("WATSONX_PASSWORD", default=None),
    )
    """Password to the CPD instance."""

    username: SecretStr | None = Field(
        alias="username",
        default_factory=secret_from_env("WATSONX_USERNAME", default=None),
    )
    """Username to the CPD instance."""

    instance_id: SecretStr | None = Field(
        alias="instance_id",
        default_factory=secret_from_env("WATSONX_INSTANCE_ID", default=None),
    )
    """Instance_id of the CPD instance."""

    version: SecretStr | None = None
    """Version of the CPD instance."""

    params: dict[str, Any] | TextChatParameters | None = None
    """Model parameters to use during request generation.

    !!! note
        `ValueError` is raised if the same Chat generation parameter is provided
        within the params attribute and as keyword argument.
    """

    frequency_penalty: float | None = None
    """Positive values penalize new tokens based on their existing frequency in the
    text so far, decreasing the model's likelihood to repeat the same line verbatim."""

    logprobs: bool | None = None
    """Whether to return log probabilities of the output tokens or not.
    If true, returns the log probabilities of each output token returned
    in the content of message."""

    top_logprobs: int | None = None
    """An integer specifying the number of most likely tokens to return at each
    token position, each with an associated log probability. The option logprobs
    must be set to true if this parameter is used."""

    max_tokens: int | None = None
    """The maximum number of tokens that can be generated in the chat completion.
    The total length of input tokens and generated tokens is limited by the
    model's context length.
    This value is now deprecated in favor of 'max_completion_tokens' parameter."""

    max_completion_tokens: int | None = None
    """The maximum number of tokens that can be generated in the chat completion.
    The total length of input tokens and generated tokens is limited by the
    model's context length."""

    n: int | None = None
    """How many chat completion choices to generate for each input message.
    Note that you will be charged based on the number of generated tokens across
    all of the choices. Keep n as 1 to minimize costs."""

    presence_penalty: float | None = None
    """Positive values penalize new tokens based on whether they appear in the
    text so far, increasing the model's likelihood to talk about new topics."""

    temperature: float | None = None
    """What sampling temperature to use. Higher values like 0.8 will make the
    output more random, while lower values like 0.2 will make it more focused
    and deterministic.

    We generally recommend altering this or top_p but not both."""

    response_format: dict[str, Any] | None = None
    """The chat response format parameters."""

    top_p: float | None = None
    """An alternative to sampling with temperature, called nucleus sampling,
    where the model considers the results of the tokens with top_p probability
    mass. So 0.1 means only the tokens comprising the top 10% probability mass
    are considered.

    We generally recommend altering this or temperature but not both."""

    time_limit: int | None = None
    """Time limit in milliseconds - if not completed within this time,
    generation will stop."""

    logit_bias: dict[str, int] | None = None
    """Increasing or decreasing probability of tokens being selected
    during generation."""

    seed: int | None = None
    """Random number generator seed to use in sampling mode
    for experimental repeatability."""

    stop: list[str] | None = None
    """Stop sequences are one or more strings which will cause the text generation
    to stop if/when they are produced as part of the output."""

    chat_template_kwargs: dict[str, Any] | None = None
    """Additional chat template parameters."""

    reasoning_effort: Literal["low", "medium", "high"] | None = None
    """A lower reasoning effort can result in faster responses, fewer tokens used,
    and shorter reasoning_content in the responses.
    Supported values are: low, medium, and high."""

    include_reasoning: bool | None = None
    """Whether to include `reasoning_content` in the response."""

    repetition_penalty: float | None = None
    """Represents the penalty for penalizing tokens that have already been generated
    or belong to the context."""

    length_penalty: float | None = None
    """Exponential penalty to the length that is used with beam-based generation.
    It is applied as an exponent to the sequence length, which in turn is used
    to divide the score of the sequence. Since the score is the log likelihood
    of the sequence (i.e. negative), `length_penalty` > 0.0 promotes longer sequences,
    while `length_penalty` < 0.0 encourages shorter sequences."""

    verify: str | bool | None = None
    """You can pass one of following as verify:
        * the path to a CA_BUNDLE file
        * the path of directory with certificates of trusted CAs
        * True - default path to truststore will be taken
        * False - no verification will be made"""

    validate_model: bool = True
    """Model ID validation."""

    streaming: bool = False
    """Whether to stream the results or not."""

    watsonx_model: ModelInference = Field(default=None, exclude=True)  #: :meta private:

    watsonx_model_gateway: Gateway = Field(
        default=None,
        exclude=True,
    )  #: :meta private:

    watsonx_client: APIClient | None = Field(default=None, exclude=True)

    model_config = ConfigDict(populate_by_name=True)

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Is lc serializable."""
        return False

    @property
    def _llm_type(self) -> str:
        """Return the type of chat model."""
        return "watsonx-chat"

    def _get_ls_params(
        self,
        stop: list[str] | None = None,
        **kwargs: Any,
    ) -> LangSmithParams:
        """Get standard params for tracing."""
        params = super()._get_ls_params(stop=stop, **kwargs)
        params["ls_provider"] = "ibm"
        if self.model_id:
            params["ls_model_name"] = self.model_id
        return params

    @property
    def lc_secrets(self) -> dict[str, str]:
        """Mapping of secret environment variables."""
        return {
            "url": "WATSONX_URL",
            "api_key": "WATSONX_API_KEY",  # preferred
            "apikey": "WATSONX_APIKEY",
            "token": "WATSONX_TOKEN",
            "password": "WATSONX_PASSWORD",
            "username": "WATSONX_USERNAME",
            "instance_id": "WATSONX_INSTANCE_ID",
        }

    @model_validator(mode="before")
    @classmethod
    def _normalize_and_warn_deprecated_input(cls, data: Any) -> Any:
        if isinstance(data, dict):
            # Handle deprecated input kwarg name `apikey` vs new `api_key`.
            data = normalize_api_key(data=data)
        return data

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that credentials and python package exists in environment."""
        self.params = self.params or {}

        if isinstance(self.params, BaseSchema):
            self.params = self.params.to_dict()

        check_duplicate_chat_params(self.params, self.__dict__)

        self.params.update(
            {
                k: v
                for k, v in {
                    param: getattr(self, param)
                    for param in ChatWatsonx._get_supported_chat_params()
                }.items()
                if v is not None
            },
        )
        if self.watsonx_model_gateway is not None:
            error_msg = (
                "Passing the 'watsonx_model_gateway' parameter to the ChatWatsonx "
                "constructor is not supported yet.",
            )
            raise NotImplementedError(error_msg)

        if isinstance(self.watsonx_model, ModelInference):
            self.model_id = self.watsonx_model.model_id
            self.deployment_id = getattr(self.watsonx_model, "deployment_id", "")
            self.project_id = self.watsonx_model._client.default_project_id  # noqa: SLF001
            self.space_id = self.watsonx_model._client.default_space_id  # noqa: SLF001
            self.params = self.watsonx_model.params
            self.watsonx_client = self.watsonx_model._client  # noqa: SLF001

        elif isinstance(self.watsonx_client, APIClient):
            if sum(map(bool, (self.model, self.model_id, self.deployment_id))) != 1:
                error_msg = (
                    "The parameters 'model', 'model_id' and 'deployment_id' are "
                    "mutually exclusive. Please specify exactly one of these "
                    "parameters when initializing ChatWatsonx.",
                )
                raise ValueError(error_msg)
            if self.model is not None:
                watsonx_model_gateway = Gateway(
                    api_client=self.watsonx_client,
                    verify=self.verify,
                )
                self.watsonx_model_gateway = watsonx_model_gateway
            else:
                watsonx_model = ModelInference(
                    model_id=self.model_id,
                    deployment_id=self.deployment_id,
                    params=self.params,
                    api_client=self.watsonx_client,
                    project_id=self.project_id,
                    space_id=self.space_id,
                    verify=self.verify,
                    validate=self.validate_model,
                )
                self.watsonx_model = watsonx_model
        else:
            if sum(map(bool, (self.model, self.model_id, self.deployment_id))) != 1:
                error_msg = (
                    "The parameters 'model', 'model_id' and 'deployment_id' are "
                    "mutually exclusive. Please specify exactly one of these "
                    "parameters when initializing ChatWatsonx.",
                )
                raise ValueError(error_msg)

            credentials = resolve_watsonx_credentials(
                url=self.url,
                api_key=self.api_key,
                token=self.token,
                password=self.password,
                username=self.username,
                instance_id=self.instance_id,
                version=self.version,
                verify=self.verify,
            )
            if self.model is not None:
                watsonx_model_gateway = Gateway(
                    credentials=credentials,
                    verify=self.verify,
                )
                self.watsonx_model_gateway = watsonx_model_gateway
            else:
                watsonx_model = ModelInference(
                    model_id=self.model_id,
                    deployment_id=self.deployment_id,
                    credentials=credentials,
                    params=self.params,
                    project_id=self.project_id,
                    space_id=self.space_id,
                    verify=self.verify,
                    validate=self.validate_model,
                )
                self.watsonx_model = watsonx_model

        return self

    @gateway_error_handler
    def _call_model_gateway(
        self, *, model: str, messages: list[dict[str, Any]], **params: Any
    ) -> Any:
        return self.watsonx_model_gateway.chat.completions.create(
            model=model,
            messages=messages,
            **params,
        )

    @async_gateway_error_handler
    async def _acall_model_gateway(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        **params: Any,
    ) -> Any:
        return await self.watsonx_model_gateway.chat.completions.acreate(
            model=model,
            messages=messages,
            **params,
        )

    def _generate(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self.streaming:
            stream_iter = self._stream(
                messages,
                stop=stop,
                run_manager=run_manager,
                **kwargs,
            )
            return generate_from_stream(stream_iter)

        message_dicts, params = self._create_message_dicts(messages, stop, **kwargs)
        updated_params = self._merge_params(params, kwargs)
        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **updated_params}
            response = self._call_model_gateway(
                model=self.model,
                messages=message_dicts,
                **call_kwargs,
            )
        else:
            response = self.watsonx_model.chat(
                messages=message_dicts,
                **(kwargs | {"params": updated_params}),
            )
        return self._create_chat_result(response)

    async def _agenerate(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self.streaming:
            stream_iter = self._astream(
                messages,
                stop=stop,
                run_manager=run_manager,
                **kwargs,
            )
            return await agenerate_from_stream(stream_iter)

        message_dicts, params = self._create_message_dicts(messages, stop, **kwargs)
        updated_params = self._merge_params(params, kwargs)
        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **updated_params}
            response = await self._acall_model_gateway(
                model=self.model,
                messages=message_dicts,
                **call_kwargs,
            )
        else:
            response = await self.watsonx_model.achat(
                messages=message_dicts,
                **(kwargs | {"params": updated_params}),
            )
        return self._create_chat_result(response)

    def _stream(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop, **kwargs)
        updated_params = self._merge_params(params, kwargs)

        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **updated_params, "stream": True}
            chunk_iter = self._call_model_gateway(
                model=self.model,
                messages=message_dicts,
                **call_kwargs,
            )
        else:
            call_kwargs = {**kwargs, "params": updated_params}
            chunk_iter = self.watsonx_model.chat_stream(
                messages=message_dicts,
                **call_kwargs,
            )

        default_chunk_class: type[BaseMessageChunk] = AIMessageChunk
        is_first_tool_chunk = True
        _prompt_tokens_included = False

        for chunk in chunk_iter:
            chunk_data = chunk if isinstance(chunk, dict) else chunk.model_dump()
            generation_chunk = _convert_chunk_to_generation_chunk(
                chunk_data,
                default_chunk_class,
                is_first_tool_chunk=is_first_tool_chunk,
                _prompt_tokens_included=_prompt_tokens_included,
            )
            if generation_chunk is None:
                continue

            if (
                hasattr(generation_chunk.message, "usage_metadata")
                and generation_chunk.message.usage_metadata
            ):
                _prompt_tokens_included = True
            default_chunk_class = generation_chunk.message.__class__
            logprobs = (generation_chunk.generation_info or {}).get("logprobs")
            if run_manager:
                run_manager.on_llm_new_token(
                    generation_chunk.text,
                    chunk=generation_chunk,
                    logprobs=logprobs,
                )
            if hasattr(generation_chunk.message, "tool_calls") and isinstance(
                generation_chunk.message.tool_calls,
                list,
            ):
                first_tool_call = (
                    generation_chunk.message.tool_calls[0]
                    if generation_chunk.message.tool_calls
                    else None
                )
                if isinstance(first_tool_call, dict) and first_tool_call.get("name"):
                    is_first_tool_chunk = False

            yield generation_chunk

    async def _astream(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop, **kwargs)
        updated_params = self._merge_params(params, kwargs)

        if self.watsonx_model_gateway is not None:
            call_kwargs = {**kwargs, **updated_params, "stream": True}
            chunk_iter = await self._acall_model_gateway(
                model=self.model,
                messages=message_dicts,
                **call_kwargs,
            )
        else:
            call_kwargs = {**kwargs, "params": updated_params}
            chunk_iter = await self.watsonx_model.achat_stream(
                messages=message_dicts,
                **call_kwargs,
            )

        default_chunk_class: type[BaseMessageChunk] = AIMessageChunk
        is_first_tool_chunk = True
        _prompt_tokens_included = False

        async for chunk in chunk_iter:
            chunk_data = chunk if isinstance(chunk, dict) else chunk.model_dump()
            generation_chunk = _convert_chunk_to_generation_chunk(
                chunk_data,
                default_chunk_class,
                is_first_tool_chunk=is_first_tool_chunk,
                _prompt_tokens_included=_prompt_tokens_included,
            )
            if generation_chunk is None:
                continue

            if (
                hasattr(generation_chunk.message, "usage_metadata")
                and generation_chunk.message.usage_metadata
            ):
                _prompt_tokens_included = True
            default_chunk_class = generation_chunk.message.__class__
            logprobs = (generation_chunk.generation_info or {}).get("logprobs")
            if run_manager:
                await run_manager.on_llm_new_token(
                    generation_chunk.text,
                    chunk=generation_chunk,
                    logprobs=logprobs,
                )
            if hasattr(generation_chunk.message, "tool_calls") and isinstance(
                generation_chunk.message.tool_calls,
                list,
            ):
                first_tool_call = (
                    generation_chunk.message.tool_calls[0]
                    if generation_chunk.message.tool_calls
                    else None
                )
                if isinstance(first_tool_call, dict) and first_tool_call.get("name"):
                    is_first_tool_chunk = False

            yield generation_chunk

    @staticmethod
    def _merge_params(params: dict[str, Any], kwargs: dict[str, Any]) -> dict[str, Any]:
        param_updates = {}
        for k in ChatWatsonx._get_supported_chat_params():
            if kwargs.get(k) is not None:
                param_updates[k] = kwargs.pop(k)

        if kwargs.get("params"):
            merged_params = merge_dicts(params, param_updates)
        else:
            merged_params = params | param_updates

        return merged_params

    def _create_message_dicts(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None,
        **kwargs: Any,
    ) -> tuple[list[dict[str, Any]], dict[str, Any]]:
        params = extract_chat_params(kwargs, self.params)

        if stop is not None:
            if params and "stop_sequences" in params:
                error_msg = (
                    "`stop_sequences` found in both the input and default params."
                )
                raise ValueError(error_msg)
            params = (params or {}) | {"stop_sequences": stop}
        message_dicts = [_convert_message_to_dict(m) for m in messages]
        return message_dicts, params or {}

    def _create_chat_result(
        self,
        response: dict[str, Any],
        generation_info: dict[str, Any] | None = None,
    ) -> ChatResult:
        generations = []

        if response.get("error"):
            raise ValueError(response.get("error"))

        token_usage = response.get("usage", {})

        for res in response["choices"]:
            message = _convert_dict_to_message(res["message"], response["id"])

            if token_usage and isinstance(message, AIMessage):
                message.usage_metadata = _create_usage_metadata(
                    token_usage, _prompt_tokens_included=False
                )
            generation_info = generation_info or {}
            generation_info["finish_reason"] = (
                res.get("finish_reason")
                if res.get("finish_reason") is not None
                else generation_info.get("finish_reason")
            )
            if "logprobs" in res:
                generation_info["logprobs"] = res["logprobs"]
            gen = ChatGeneration(message=message, generation_info=generation_info)
            generations.append(gen)
        llm_output = {
            "token_usage": token_usage,
            "model_name": response.get("model_id", self.model_id),
            "system_fingerprint": response.get("system_fingerprint", ""),
        }

        return ChatResult(generations=generations, llm_output=llm_output)

    @staticmethod
    def _get_supported_chat_params() -> list[str]:
        # watsonx.ai Chat API doc: https://cloud.ibm.com/apidocs/watsonx-ai#text-chat
        return [
            "frequency_penalty",
            "logprobs",
            "top_logprobs",
            "max_tokens",
            "max_completion_tokens",
            "n",
            "presence_penalty",
            "response_format",
            "temperature",
            "top_p",
            "time_limit",
            "logit_bias",
            "seed",
            "stop",
            "chat_template_kwargs",
            "reasoning_effort",
            "include_reasoning",
            "repetition_penalty",
            "length_penalty",
        ]

    def bind_tools(
        self,
        tools: Sequence[dict[str, Any] | type | Callable[..., Any] | BaseTool],
        *,
        tool_choice: dict[str, Any] | str | bool | None = None,
        strict: bool | None = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, AIMessage]:
        """Bind tool-like objects to this chat model.

        Args:
            tools: A list of tool definitions to bind to this chat model.
                Can be  a dictionary, pydantic model, callable, or BaseTool. Pydantic
                models, callables, and BaseTools will be automatically converted to
                their schema dictionary representation.
            tool_choice: Which tool to require the model to call. Options are:

                - `str` of the form `'<<tool_name>>'`: calls `<<tool_name>>` tool.
                - `'auto'`: automatically selects a tool (including no tool).
                - `'none'`: does not call a tool.
                - `'any'` or `'required'` or `True`: force at least one tool to be
                  called.
                - `dict` of the form
                  `{"type": "function", "function": {"name": <<tool_name>>}}`:
                  calls `<<tool_name>>` tool.
                - `False` or `None`: no effect, default OpenAI behavior.

            strict: If `True`, model output is guaranteed to exactly match the JSON
                Schema provided in the tool definition.
                The input schema will also be validated according to the supported
                schemas.
                If `False`, input schema will not be validated and model output will not
                be validated.
                If `None`, `strict` argument will not be passed to the model.

            kwargs: Any additional parameters are passed directly to `bind`.
        """
        formatted_tools = [
            convert_to_openai_tool(tool, strict=strict) for tool in tools
        ]
        if tool_choice:
            if isinstance(tool_choice, str):
                # tool_choice is a tool/function name
                if tool_choice not in ("auto", "none", "any", "required"):
                    tool_choice = {
                        "type": "function",
                        "function": {"name": tool_choice},
                    }
                # We support 'any' since other models use this instead of 'required'.
                if tool_choice == "any":
                    tool_choice = "required"
            elif isinstance(tool_choice, bool):
                tool_choice = "required"
            elif isinstance(tool_choice, dict):
                tool_names = [
                    formatted_tool["function"]["name"]
                    for formatted_tool in formatted_tools
                ]
                if not any(
                    tool_name == tool_choice["function"]["name"]
                    for tool_name in tool_names
                ):
                    error_msg = (
                        f"Tool choice {tool_choice} was specified, but the only "
                        f"provided tools were {tool_names}.",
                    )
                    raise ValueError(error_msg)
            else:
                error_msg = (  # type: ignore[unreachable]
                    f"Unrecognized tool_choice type. Expected str, bool or dict. "
                    f"Received: {tool_choice}",
                )
                raise ValueError(error_msg)

            if isinstance(tool_choice, str):
                kwargs["tool_choice_option"] = tool_choice
            else:
                kwargs["tool_choice"] = tool_choice
        else:
            kwargs["tool_choice_option"] = "auto"

        return super().bind(tools=formatted_tools, **kwargs)

    @override
    def with_structured_output(
        self,
        schema: dict[str, Any] | type | None = None,
        *,
        method: Literal[
            "function_calling",
            "json_mode",
            "json_schema",
        ] = "function_calling",
        include_raw: bool = False,
        strict: bool | None = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, dict[str, Any] | BaseModel]:
        r"""Model wrapper that returns outputs formatted to match the given schema.

        Args:
            schema: The output schema. Can be passed in as:

                - a JSON Schema,
                - a TypedDict class,
                - or a Pydantic class,

                If `schema` is a Pydantic class then the model output will be a
                Pydantic instance of that class, and the model-generated fields will be
                validated by the Pydantic class. Otherwise the model output will be a
                dict and will not be validated.

            method: The method for steering model generation, one of:

                - `'function_calling'`: uses tool-calling features.
                - `'json_schema'`: uses dedicated structured output features.
                - `'json_mode'`: uses JSON mode.

            include_raw:
                If False then only the parsed structured output is returned. If
                an error occurs during model output parsing it will be raised. If True
                then both the raw model response (a BaseMessage) and the parsed model
                response will be returned. If an error occurs during output parsing it
                will be caught and returned as well. The final output is always a dict
                with keys `'raw'`, `'parsed'`, and `'parsing_error'`.

            strict:

                - `True`:
                    Model output is guaranteed to exactly match the schema.
                    The input schema will also be validated according to the
                    supported schemas.
                - `False`:
                    Input schema will not be validated and model output will not be
                    validated.
                - `None`:
                    `strict` argument will not be passed to the model.

            kwargs: Additional keyword args


        Returns:
            A Runnable that takes same inputs as a `langchain_core.language_models.chat.BaseChatModel`.

            If `include_raw` is True, then Runnable outputs a dict with keys:

            - `'raw'`: BaseMessage
            - `'parsed'`: None if there was a parsing error, otherwise the type depends on the `schema` as described above.
            - `'parsing_error'`: Optional[BaseException]

        ??? note "Example: `schema=Pydantic` class, `method='function_calling'`, `include_raw=True`"

            ```python
            from langchain_ibm import ChatWatsonx
            from pydantic import BaseModel


            class AnswerWithJustification(BaseModel):
                '''An answer to the user question along with justification for the answer.'''

                answer: str
                justification: str


            model = ChatWatsonx(...)
            structured_model = model.with_structured_output(
                AnswerWithJustification, include_raw=True
            )

            structured_model.invoke(
                "What weighs more a pound of bricks or a pound of feathers"
            )
            ```

            ```python
            {
                "raw": AIMessage(
                    content="",
                    additional_kwargs={
                        "tool_calls": [
                            {
                                "id": "call_Ao02pnFYXD6GN1yzc0uXPsvF",
                                "function": {
                                    "arguments": '{"answer":"They weigh the same.","justification":"Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume or density of the objects may differ."}',
                                    "name": "AnswerWithJustification",
                                },
                                "type": "function",
                            }
                        ]
                    },
                ),
                "parsed": AnswerWithJustification(
                    answer="They weigh the same.",
                    justification="Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume or density of the objects may differ.",
                ),
                "parsing_error": None,
            }
            ```

        ??? note "Example: `schema=JSON` schema, `method='function_calling'`, `include_raw=False`"

            ```python
            from langchain_ibm import ChatWatsonx
            from pydantic import BaseModel
            from langchain_core.utils.function_calling import convert_to_openai_tool


            class AnswerWithJustification(BaseModel):
                '''An answer to the user question along with justification for the answer.'''

                answer: str
                justification: str


            dict_schema = convert_to_openai_tool(AnswerWithJustification)
            model = ChatWatsonx(...)
            structured_model = model.with_structured_output(dict_schema)

            structured_model.invoke(
                "What weighs more a pound of bricks or a pound of feathers"
            )
            ```

            ```python
            {
                "answer": "They weigh the same",
                "justification": "Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume and density of the two substances differ.",
            }
            ```

        ??? note "Example: `schema=Pydantic` class, `method='json_schema'`, `include_raw=True`"

            ```python
            from langchain_ibm import ChatWatsonx
            from pydantic import BaseModel


            class AnswerWithJustification(BaseModel):
                '''An answer to the user question along with justification for the answer.'''

                answer: str
                justification: str


            model = ChatWatsonx(...)
            structured_model = model.with_structured_output(
                AnswerWithJustification, method="json_schema", include_raw=True
            )

            structured_model.invoke(
                "What weighs more a pound of bricks or a pound of feathers"
            )
            ```

            ```python
            {
                "raw": AIMessage(
                    content="",
                    additional_kwargs={
                        "tool_calls": [
                            {
                                "id": "chatcmpl-tool-bfbd6f6dd33b438990c5ddf277485971",
                                "type": "function",
                                "function": {
                                    "name": "AnswerWithJustification",
                                    "arguments": '{"answer": "They weigh the same", "justification": "A pound is a unit of weight or mass, so both a pound of bricks and a pound of feathers weigh the same amount, one pound."}',
                                },
                            }
                        ]
                    },
                    response_metadata={
                        "token_usage": {
                            "completion_tokens": 45,
                            "prompt_tokens": 275,
                            "total_tokens": 320,
                        },
                        "model_name": "meta-llama/llama-3-3-70b-instruct",
                        "system_fingerprint": "",
                        "finish_reason": "stop",
                    },
                    id="chatcmpl-461ca5bd-1982-412c-b886-017c483bf481---8c18b06eead65ae4691364798787bda7---71896588-efa5-439f-a25f-d1abfe289f5a",
                    tool_calls=[
                        {
                            "name": "AnswerWithJustification",
                            "args": {
                                "answer": "They weigh the same",
                                "justification": "A pound is a unit of weight or mass, so both a pound of bricks and a pound of feathers weigh the same amount, one pound.",
                            },
                            "id": "chatcmpl-tool-bfbd6f6dd33b438990c5ddf277485971",
                            "type": "tool_call",
                        }
                    ],
                    usage_metadata={
                        "input_tokens": 275,
                        "output_tokens": 45,
                        "total_tokens": 320,
                    },
                ),
                "parsed": AnswerWithJustification(
                    answer="They weigh the same",
                    justification="A pound is a unit of weight or mass, so both a pound of bricks and a pound of feathers weigh the same amount, one pound.",
                ),
                "parsing_error": None,
            }
            ```

        ??? note "Example: `schema=function` schema, `method='json_schema'`, `include_raw=False`"

            ```python
            from langchain_ibm import ChatWatsonx
            from pydantic import BaseModel

            function__schema = {
                "name": "AnswerWithJustification",
                "description": "An answer to the user question along with justification for the answer.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "answer": {"type": "string"},
                        "justification": {
                            "description": "A justification for the answer.",
                            "type": "string",
                        },
                    },
                    "required": ["answer"],
                },
            }

            model = ChatWatsonx(...)
            structured_model = model.with_structured_output(
                function_schema, method="json_schema", include_raw=False
            )

            structured_model.invoke(
                "What weighs more a pound of bricks or a pound of feathers"
            )
            ```

            ```python
            {
                "answer": "They weigh the same",
                "justification": "Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume and density of the two substances differ.",
            }
            ```

        ??? note "Example: `schema=Pydantic` class, `method='json_mode'`, `include_raw=True`"

            ```python
            from langchain_ibm import ChatWatsonx
            from pydantic import BaseModel


            class AnswerWithJustification(BaseModel):
                answer: str
                justification: str


            model = ChatWatsonx(...)
            structured_model = model.with_structured_output(
                AnswerWithJustification, method="json_mode", include_raw=True
            )

            structured_model.invoke(
                "Answer the following question. "
                "Make sure to return a JSON blob with keys 'answer' and 'justification'.\n\n"
                "What's heavier a pound of bricks or a pound of feathers?"
            )
            ```

            ```python
            {
                "raw": AIMessage(
                    content='{\n    "answer": "They are both the same weight.",\n    "justification": "Both a pound of bricks and a pound of feathers weigh one pound. The difference lies in the volume and density of the materials, not the weight." \n}'
                ),
                "parsed": AnswerWithJustification(
                    answer="They are both the same weight.",
                    justification="Both a pound of bricks and a pound of feathers weigh one pound. The difference lies in the volume and density of the materials, not the weight.",
                ),
                "parsing_error": None,
            }
            ```

        ??? note "Example: `schema=None`, `method='json_mode'`, `include_raw=True`"

            ```python
            from langchain_ibm import ChatWatsonx

            model = ChatWatsonx(...)
            structured_model = model.with_structured_output(
                method="json_mode", include_raw=True
            )

            structured_model.invoke(
                "Answer the following question. "
                "Make sure to return a JSON blob with keys 'answer' and 'justification'.\n\n"
                "What's heavier a pound of bricks or a pound of feathers?"
            )
            ```

            ```python
            {
                "raw": AIMessage(
                    content='{\n    "answer": "They are both the same weight.",\n    "justification": "Both a pound of bricks and a pound of feathers weigh one pound. The difference lies in the volume and density of the materials, not the weight." \n}'
                ),
                "parsed": {
                    "answer": "They are both the same weight.",
                    "justification": "Both a pound of bricks and a pound of feathers weigh one pound. The difference lies in the volume and density of the materials, not the weight.",
                },
                "parsing_error": None,
            }
            ```

        """  # noqa: E501
        if strict is not None and method == "json_mode":
            msg = "Argument `strict` is not supported with `method`='json_mode'"
            raise ValueError(msg)

        is_pydantic_schema = _is_pydantic_class(schema)

        if (
            method == "json_schema"
            and is_pydantic_schema
            and issubclass(schema, BaseModelV1)  # type: ignore[arg-type]
        ):
            # Check for Pydantic BaseModel V1
            warnings.warn(
                "Received a Pydantic BaseModel V1 schema. This is not supported by "
                'method="json_schema". Please use method="function_calling" '
                "or specify schema via JSON Schema or Pydantic V2 BaseModel. "
                'Overriding to method="function_calling".',
                stacklevel=2,
            )
            method = "function_calling"

        is_pydantic_schema = _is_pydantic_class(schema)
        if method == "function_calling":
            if schema is None:
                error_msg = (
                    "schema must be specified when method is not 'json_mode'. "
                    "Received None."
                )
                raise ValueError(error_msg)
            formatted_tool = convert_to_openai_tool(schema)
            tool_name = formatted_tool["function"]["name"]
            model = self.bind_tools(
                [schema],
                strict=strict,
                tool_choice=tool_name,
                ls_structured_output_format={
                    "kwargs": {"method": method, "strict": strict},
                    "schema": formatted_tool,
                },
            )
            if is_pydantic_schema:
                output_parser: Runnable[Any, Any] = PydanticToolsParser(
                    tools=[schema],
                    first_tool_only=True,
                )
            else:
                output_parser = JsonOutputKeyToolsParser(
                    key_name=tool_name,
                    first_tool_only=True,
                )
        elif method == "json_mode":
            model = self.bind(
                response_format={"type": "json_object"},
                ls_structured_output_format={
                    "kwargs": {"method": method},
                    "schema": schema,
                },
            )
            output_parser = (
                PydanticOutputParser(pydantic_object=schema)  # type: ignore[unused-ignore]
                if is_pydantic_schema
                else JsonOutputParser()
            )
        elif method == "json_schema":
            if schema is None:
                error_msg = (
                    "schema must be specified when method is not 'json_mode'. "
                    "Received None."
                )
                raise ValueError(error_msg)
            if is_pydantic_schema:
                response_format = {
                    "type": "json_schema",
                    "json_schema": {
                        "name": schema.__name__,  # type: ignore[union-attr]
                        "description": schema.__doc__,
                        "schema": schema.model_json_schema(),  # type: ignore[union-attr]
                    },
                }
            else:
                response_format = _convert_to_openai_response_format(  # type: ignore[assignment]
                    schema, strict=strict
                )

            bind_kwargs = {
                "response_format": response_format,
                "ls_structured_output_format": {
                    "kwargs": {"method": method, "strict": strict},
                    "schema": convert_to_openai_tool(schema),
                },
            }
            model = self.bind(**bind_kwargs)
            output_parser = (
                PydanticOutputParser(pydantic_object=schema)  # type: ignore[unused-ignore]
                if is_pydantic_schema
                else JsonOutputParser()
            )
        else:
            error_msg = (  # type: ignore[unreachable]
                f"Unrecognized method argument. Expected one of 'function_calling' or "
                f"'json_mode'. Received: '{method}'"
            )
            raise ValueError(error_msg)

        if include_raw:
            parser_assign = RunnablePassthrough.assign(
                parsed=itemgetter("raw") | output_parser,
                parsing_error=lambda _: None,
            )
            parser_none = RunnablePassthrough.assign(parsed=lambda _: None)
            parser_with_fallback = parser_assign.with_fallbacks(
                [parser_none],
                exception_key="parsing_error",
            )
            return RunnableMap(raw=model) | parser_with_fallback
        return model | output_parser


def _is_pydantic_class(obj: Any) -> bool:
    return isinstance(obj, type) and is_basemodel_subclass(obj)


def _lc_tool_call_to_watsonx_tool_call(tool_call: ToolCall) -> dict[str, Any]:
    return {
        "type": "function",
        "id": tool_call["id"],
        "function": {
            "name": tool_call["name"],
            "arguments": json.dumps(tool_call["args"], ensure_ascii=False),
        },
    }


def _lc_invalid_tool_call_to_watsonx_tool_call(
    invalid_tool_call: InvalidToolCall,
) -> dict[str, Any]:
    return {
        "type": "function",
        "id": invalid_tool_call["id"],
        "function": {
            "name": invalid_tool_call["name"],
            "arguments": invalid_tool_call["args"],
        },
    }


def _create_usage_metadata(
    oai_token_usage: dict[str, Any],
    *,
    _prompt_tokens_included: bool,
) -> UsageMetadata:
    input_tokens = (
        oai_token_usage.get("prompt_tokens", 0) if not _prompt_tokens_included else 0
    )
    output_tokens = oai_token_usage.get("completion_tokens", 0)
    total_tokens = oai_token_usage.get("total_tokens", input_tokens + output_tokens)
    return UsageMetadata(
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        total_tokens=total_tokens,
    )


def _convert_to_openai_response_format(
    schema: dict[str, Any] | type, *, strict: bool | None = None
) -> dict[str, Any] | TypeBaseModel:
    if isinstance(schema, type) and is_basemodel_subclass(schema):
        return schema

    if (
        isinstance(schema, dict)
        and "json_schema" in schema
        and schema.get("type") == "json_schema"
    ):
        response_format = schema
    elif isinstance(schema, dict) and "name" in schema and "schema" in schema:
        response_format = {"type": "json_schema", "json_schema": schema}
    else:
        if strict is None:
            if isinstance(schema, dict) and isinstance(schema.get("strict"), bool):
                strict = schema["strict"]
            else:
                strict = False
        function = convert_to_openai_function(schema, strict=strict)
        function["schema"] = function.pop("parameters")
        response_format = {"type": "json_schema", "json_schema": function}

    if (
        strict is not None
        and strict is not response_format["json_schema"].get("strict")
        and isinstance(schema, dict)
    ):
        msg = (
            f"Output schema already has 'strict' value set to "
            f"{schema['json_schema']['strict']} but 'strict' also passed in to "
            f"with_structured_output as {strict}. Please make sure that "
            f"'strict' is only specified in one place."
        )
        raise ValueError(msg)
    return response_format



"""IBM watsonx.ai Toolkit wrapper."""

from typing import Any, cast

from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models.utils import (  # type: ignore[import-untyped]
    Tool,
    Toolkit,
)
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.tools.base import BaseTool, BaseToolkit
from langchain_core.utils.utils import secret_from_env
from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    SecretStr,
    create_model,
    model_validator,
)
from typing_extensions import Self, override

from langchain_ibm.agent_toolkits.utility.utils import convert_to_watsonx_tool
from langchain_ibm.utils import (
    normalize_api_key,
    resolve_watsonx_credentials,
    secret_from_env_multi,
)


class WatsonxTool(BaseTool):
    """IBM watsonx.ai Tool."""

    name: str
    """Name of the tool."""

    description: str
    """Description of what the tool is used for."""

    agent_description: str | None = None
    """The precise instruction to agent LLMs
    and should be treated as part of the system prompt."""

    tool_input_schema: dict[str, Any] | None = None
    """Schema of the input that is provided when running the tool if applicable."""

    tool_config_schema: dict[str, Any] | None = None
    """Schema of the config that can be provided when running the tool if applicable."""

    tool_config: dict[str, Any] | None = None
    """Config properties to be used when running a tool if applicable."""

    args_schema: type[BaseModel] = BaseModel

    _watsonx_tool: Tool  #: :meta private:

    watsonx_client: APIClient = Field(exclude=True)

    @model_validator(mode="after")
    def validate_tool(self) -> Self:
        """Validate tool."""
        self._watsonx_tool = Tool(
            api_client=self.watsonx_client,
            name=self.name,
            description=self.description,
            agent_description=self.agent_description,
            input_schema=self.tool_input_schema,
            config_schema=self.tool_config_schema,
        )
        converted_tool = convert_to_watsonx_tool(self)
        json_schema = converted_tool["function"]["parameters"]
        self.args_schema = _json_schema_to_pydantic_model(
            name="ToolArgsSchema",
            schema=json_schema,
        )

        return self

    @override
    def _run(
        self,
        *args: Any,
        run_manager: CallbackManagerForToolRun | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run the tool."""
        if self.tool_input_schema is None:
            input_data = kwargs.get("input") or args[0]
        else:
            input_data = {
                k: v
                for k, v in kwargs.items()
                if k in self.tool_input_schema["properties"]
            }

        return self._watsonx_tool.run(input_data, self.tool_config)

    def set_tool_config(self, tool_config: dict[str, Any]) -> None:
        """Set tool config properties.

        ???+ example "Example"

            ```python
            google_search = watsonx_toolkit.get_tool("GoogleSearch")
            print(google_search.tool_config_schema)
            tool_config = {"maxResults": 3}
            google_search.set_tool_config(tool_config)
            ```

        """
        self.tool_config = tool_config


class WatsonxToolkit(BaseToolkit):
    """IBM watsonx.ai Toolkit.

    ???+ info "Setup"

        To use, you should have `langchain_ibm` python package installed,
        and the environment variable `WATSONX_API_KEY` set with your API key, or pass
        it as a named parameter `api_key` to the constructor.

        ```bash
        pip install -U langchain-ibm

        # or using uv
        uv add langchain-ibm
        ```

        ```bash
        export WATSONX_API_KEY="your-api-key"
        ```

        !!! deprecated
            `apikey` and `WATSONX_APIKEY` are deprecated and will be removed in
            version `2.0.0`. Use `api_key` and `WATSONX_API_KEY` instead.

    ??? info "Instantiate"

        IBM watsonx.ai for IBM Cloud:

        ```python
        from langchain_ibm.agent_toolkits.utility import WatsonxToolkit

        watsonx_toolkit = WatsonxToolkit(
            url="https://us-south.ml.cloud.ibm.com",
            project_id="*****",  # or `space_id`
            api_key="*****",  # not needed if `WATSONX_API_KEY` is set
        )
        ```

        IBM watsonx.ai software:
        ```python
        from langchain_ibm.agent_toolkits.utility import WatsonxToolkit

        watsonx_toolkit = WatsonxToolkit(
            url="<CPD_URL>",
            project_id="*****",  # or `space_id`
            username="*****",
            password="*****",
            instance_id="*****",
            version="*****",  # optional
        )
        ```

    ??? info "Invoke"

        ```python
        tools = watsonx_toolkit.get_tools()

        google_search = watsonx_toolkit.get_tool(tool_name="GoogleSearch")

        tool_config = {
            "maxResults": 3,
        }
        google_search.set_tool_config(tool_config)
        input = {
            "input": "Search IBM",
        }
        search_result = google_search.invoke(input)
        ```

    ??? info "Run"

        ```python
        rag_query = watsonx_toolkit.get_tool(tool_name="RAGQuery")

        rag_query.set_tool_config(
            {
                "vectorIndexId": "<vector-index-id>",
                "projectId": "<project-id>",
            }
        )

        res = rag_query.run("How to initialize APIClient?")
        ```

    """

    project_id: str | None = None
    """ID of the watsonx.ai Studio project."""

    space_id: str | None = None
    """ID of the watsonx.ai Studio space."""

    url: SecretStr = Field(
        alias="url",
        default_factory=secret_from_env("WATSONX_URL", default=None),  # type: ignore[assignment]
    )
    """URL to the watsonx.ai Runtime."""

    apikey: SecretStr | None = None
    api_key: SecretStr | None = Field(
        default_factory=secret_from_env_multi(
            names_priority=["WATSONX_API_KEY", "WATSONX_APIKEY"],
            deprecated={"WATSONX_APIKEY"},
        ),
        serialization_alias="api_key",
        validation_alias=AliasChoices("api_key", "apikey"),  # accept both on input
        description="API key to the Watson Machine Learning or CPD instance.",
    )
    """API key to the watsonx.ai Runtime."""

    token: SecretStr | None = Field(
        alias="token",
        default_factory=secret_from_env("WATSONX_TOKEN", default=None),
    )
    """Token to the watsonx.ai Runtime."""

    password: SecretStr | None = Field(
        alias="password",
        default_factory=secret_from_env("WATSONX_PASSWORD", default=None),
    )
    """Password to the CPD instance."""

    username: SecretStr | None = Field(
        alias="username",
        default_factory=secret_from_env("WATSONX_USERNAME", default=None),
    )
    """Username to the CPD instance."""

    instance_id: SecretStr | None = Field(
        alias="instance_id",
        default_factory=secret_from_env("WATSONX_INSTANCE_ID", default=None),
    )
    """Instance_id of the CPD instance."""

    version: SecretStr | None = None
    """Version of the CPD instance."""

    verify: str | bool | None = None
    """You can pass one of following as verify:
        * the path to a CA_BUNDLE file
        * the path of directory with certificates of trusted CAs
        * True - default path to truststore will be taken
        * False - no verification will be made"""

    _tools: list[WatsonxTool] | None = None
    """Tools in the toolkit."""

    _watsonx_toolkit: Toolkit | None = PrivateAttr(default=None)  #: :meta private:

    watsonx_client: APIClient | None = Field(default=None, exclude=True)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="before")
    @classmethod
    def _normalize_and_warn_deprecated_input(cls, data: Any) -> Any:
        if isinstance(data, dict):
            # Handle deprecated input kwarg name `apikey` vs new `api_key`.
            data = normalize_api_key(data=data)
        return data

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that credentials and python package exists in environment."""
        if isinstance(self.watsonx_client, APIClient):
            self._watsonx_toolkit = Toolkit(self.watsonx_client)
        else:
            credentials = resolve_watsonx_credentials(
                url=self.url,
                api_key=self.api_key,
                token=self.token,
                password=self.password,
                username=self.username,
                instance_id=self.instance_id,
                version=self.version,
                verify=self.verify,
            )
            self.watsonx_client = APIClient(
                credentials=credentials,
                project_id=self.project_id,
                space_id=self.space_id,
            )
            self._watsonx_toolkit = Toolkit(self.watsonx_client)

        self._tools = [
            WatsonxTool(
                watsonx_client=self.watsonx_client,
                name=tool["name"],
                description=tool["description"],
                agent_description=tool.get("agent_description"),
                tool_input_schema=tool.get("input_schema"),
                tool_config_schema=tool.get("config_schema"),
            )
            for tool in self._watsonx_toolkit.get_tools()
        ]

        return self

    def get_tools(self) -> list[WatsonxTool]:  # type: ignore[override]
        """Get the tools in the toolkit."""
        return self._tools  # type: ignore[return-value]

    def get_tool(self, tool_name: str) -> WatsonxTool:
        """Get the tool with a given name."""
        for tool in self.get_tools():
            if tool.name == tool_name:
                return tool
        error_msg = f"A tool with the given name ({tool_name}) was not found."
        raise ValueError(error_msg)


def _json_schema_to_pydantic_model(
    name: str,
    schema: dict[str, Any],
) -> type[BaseModel]:
    properties = schema.get("properties", {})
    fields = {}

    type_mapping = {
        "string": str,
        "integer": int,
        "number": float,
        "boolean": bool,
        "array": list,
        "object": dict,
    }

    for field_name, field_schema in properties.items():
        field_type = field_schema.get("type", "string")
        is_required = field_name in schema.get("required", [])

        py_type = type_mapping.get(field_type, Any)

        fields[field_name] = (py_type, ... if is_required else None)

    return cast("type[BaseModel]", create_model(name, **fields))  # type: ignore[call-overload]



import os

import pytest
from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]

from langchain_ibm import WatsonxLLM

URL = "https://us-south.ml.cloud.ibm.com"
WX_APIKEY = os.environ.get("WATSONX_APIKEY", "")
WX_PROJECT_ID = os.environ.get("WATSONX_PROJECT_ID", "")

MODEL = "ibm/granite-3-8b-instruct"


@pytest.mark.skip("Until not released on us-south, should be skipped.")
class TestLLMGateway:
    prompt = "Hello, How are you!"

    def test_llm_model_gateway_init_credentials(self) -> None:
        chat = WatsonxLLM(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        assert isinstance(chat, WatsonxLLM)

        response = chat.invoke(self.prompt)

        assert response
        assert isinstance(response, str)

    def test_llm_model_gateway_init_client(self) -> None:
        credentials = Credentials(url=URL, api_key=WX_APIKEY)
        api_client = APIClient(credentials=credentials, project_id=WX_PROJECT_ID)

        chat = WatsonxLLM(model=MODEL, watsonx_client=api_client)
        assert isinstance(chat, WatsonxLLM)

        response = chat.invoke(self.prompt)

        assert response
        assert isinstance(response, str)

    def test_llm_model_gateway_invoke(self) -> None:
        chat = WatsonxLLM(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.invoke(self.prompt)
        assert response
        assert isinstance(response, str)

    async def test_llm_model_gateway_ainvoke(self) -> None:
        chat = WatsonxLLM(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = await chat.ainvoke(self.prompt)
        assert response
        assert isinstance(response, str)

    def test_llm_model_gateway_generate(self) -> None:
        chat = WatsonxLLM(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.generate(["What color sunflower is?"])
        assert response
        for generation in response.generations:
            assert generation[0].text

    async def test_llm_model_gateway_agenerate(self) -> None:
        chat = WatsonxLLM(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = await chat.agenerate(["What color sunflower is?"])
        assert response
        for generation in response.generations:
            assert generation[0].text

    def test_chat_model_gateway_stream(self) -> None:
        chat = WatsonxLLM(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.stream(self.prompt)
        for chunk in response:
            assert isinstance(chunk, str)

    async def test_chat_model_gateway_astream(self) -> None:
        chat = WatsonxLLM(
            model=MODEL,
            url=URL,
            apikey=WX_APIKEY,
            project_id=WX_PROJECT_ID,
        )
        response = chat.astream(self.prompt)
        async for chunk in response:
            assert isinstance(chunk, str)



"""Test WatsonxLLM API wrapper."""

import os
import re
from unittest.mock import Mock

import pytest
from ibm_watsonx_ai import APIClient  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models import (  # type: ignore[import-untyped]
    Model,
    ModelInference,
)
from ibm_watsonx_ai.gateway import Gateway  # type: ignore[import-untyped]
from ibm_watsonx_ai.wml_client_error import (  # type: ignore[import-untyped]
    WMLClientError,
)

from langchain_ibm import WatsonxLLM

os.environ.pop("WATSONX_APIKEY", None)
os.environ.pop("WATSONX_PROJECT_ID", None)

MODEL_ID = "mistralai/mixtral-8x7b-instruct-v01"

api_client_mock = Mock(spec=APIClient)
api_client_mock.default_space_id = None
api_client_mock.default_project_id = "fake_project_id"

gateway_mock = Mock(spec=Gateway)
gateway_mock._client = api_client_mock

model_inference_mock = Mock(spec=ModelInference)
model_inference_mock._client = api_client_mock
model_inference_mock.model_id = "fake_model_id"
model_inference_mock.params = {"temperature": 1}

model_mock = Mock(spec=Model)
model_mock._client = api_client_mock
model_mock.model_id = "fake_model_id"
model_mock.params = {"temperature": 1}


def test_initialize_watsonxllm_without_url() -> None:
    pattern = r"(?=.*url)(?=.*WATSONX_URL)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(
            model_id="google/flan-ul2",
        )


def test_initialize_watsonxllm_cloud_only_url() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(model_id="google/flan-ul2", url="https://us-south.ml.cloud.ibm.com")


def test_initialize_watsonxllm_with_deprecated_apikey() -> None:
    with (
        pytest.warns(
            DeprecationWarning,
            match="'apikey' parameter is deprecated; use 'api_key' instead.",
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxLLM(
            model_id=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonxllm_with_api_key_and_apikey() -> None:
    with (
        pytest.warns(
            UserWarning,
            match="Both 'api_key' and deprecated 'apikey' were provided; "
            "'api_key' takes precedence.",
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxLLM(
            model_id=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="fake_apikey",
            api_key="fake_api_key",
        )


def test_initialize_watsonxllm_cpd_without_all() -> None:
    pattern = (
        r"(?=.*api_key)(?=.*password)(?=.*token)"
        r"(?=.*WATSONX_API_KEY)(?=.*WATSONX_PASSWORD)(?=.*WATSONX_TOKEN)"
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(
            model_id="google/flan-ul2",
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
        )


def test_initialize_watsonxllm_cpd_only_password() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(
            model_id="google/flan-ul2",
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            password="test_password",  # noqa: S106
        )


def test_initialize_watsonxllm_cpd_only_apikey() -> None:
    pattern = r"(?=.*username)(?=.*WATSONX_USERNAME)"
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(
            model_id="google/flan-ul2",
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonxllm_cpd_deprecation_warning_with_instance_id() -> None:
    with (
        pytest.warns(
            DeprecationWarning, match="The `instance_id` parameter is deprecated"
        ),
        pytest.raises(WMLClientError),
    ):
        WatsonxLLM(
            model_id="google/flan-ul2",
            url="https://cpd-zen.apps.cpd48.cp.fyre.ibm.com",
            apikey="test_apikey",
            username="test_user",
            instance_id="openshift",
        )


def test_initialize_watsonxllm_with_two_exclusive_parameters() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing WatsonxLLM."
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(
            model_id=MODEL_ID,
            model=MODEL_ID,
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonxllm_with_three_exclusive_parameters() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing WatsonxLLM."
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(
            model_id=MODEL_ID,
            model=MODEL_ID,
            deployment_id="test_deployment_id",
            url="https://us-south.ml.cloud.ibm.com",
            apikey="test_apikey",
        )


def test_initialize_watsonxllm_with_api_client_only() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing WatsonxLLM."
    )
    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM(watsonx_client=api_client_mock)


def test_initialize_watsonxllm_with_watsonx_model_gateway() -> None:
    pattern = re.escape(
        "Passing the 'watsonx_model_gateway' parameter to the WatsonxLLM "
        "constructor is not supported yet."
    )

    with pytest.raises(NotImplementedError, match=pattern):
        WatsonxLLM(watsonx_model_gateway=gateway_mock)


def test_initialize_watsonxllm_without_any_params() -> None:
    pattern = re.escape(
        "The parameters 'model', 'model_id' and 'deployment_id' are mutually exclusive."
        " Please specify exactly one of these parameters when initializing WatsonxLLM."
    )

    with pytest.raises(ValueError, match=pattern):
        WatsonxLLM()


def test_initialize_watsonxllm_with_model_inference_only() -> None:
    chat = WatsonxLLM(watsonx_model=model_inference_mock)

    assert isinstance(chat, WatsonxLLM)


def test_initialize_watsonxllm_with_model_only() -> None:
    chat = WatsonxLLM(watsonx_model=model_mock)

    assert isinstance(chat, WatsonxLLM)



"""Utility helpers."""

from copy import deepcopy
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from langchain_ibm.agent_toolkits.utility.toolkit import WatsonxTool


def convert_to_watsonx_tool(tool: "WatsonxTool") -> dict[str, Any]:
    """Convert `WatsonxTool` to watsonx tool structure.

    Args:
        tool: `WatsonxTool` from `WatsonxToolkit`

    ???+ example "Example"

        ```python
        from langchain_ibm.agents_toolkits.utility import WatsonxToolkit

        watsonx_toolkit = WatsonxToolkit(
            url="https://us-south.ml.cloud.ibm.com",
            api_key="*****",
        )
        weather_tool = watsonx_toolkit.get_tool("Weather")
        convert_to_watsonx_tool(weather_tool)
        ```

        ```json
        {
            "type": "function",
            "function": {
                "name": "Weather",
                "description": "Find the weather for a city.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "title": "location",
                            "description": "Name of the location",
                            "type": "string",
                        },
                        "country": {
                            "title": "country",
                            "description": "Name of the state or country",
                            "type": "string",
                        },
                    },
                    "required": ["location"],
                },
            },
        }
        ```
    """

    def parse_parameters(input_schema: dict[str, Any] | None) -> dict[str, Any]:
        if input_schema:
            parameters = deepcopy(input_schema)
        else:
            parameters = {
                "type": "object",
                "properties": {
                    "input": {
                        "description": "Input to be used when running tool.",
                        "type": "string",
                    },
                },
                "required": ["input"],
            }

        return parameters

    return {
        "type": "function",
        "function": {
            "name": tool.name,
            "description": tool.description,
            "parameters": parse_parameters(tool.tool_input_schema),
        },
    }



"""Utility helpers for langchain-ibm."""

import functools
import logging
import os
import warnings
from collections.abc import Callable
from copy import deepcopy
from typing import Any, cast
from urllib.parse import urlparse

from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]
from ibm_watsonx_ai.foundation_models.schema import (  # type: ignore[import-untyped]
    BaseSchema,
)
from ibm_watsonx_ai.wml_client_error import (  # type: ignore[import-untyped]
    ApiRequestFailure,
)
from pydantic import SecretStr

logger = logging.getLogger(__name__)


def check_for_attribute(value: SecretStr | None, key: str, env_key: str) -> None:
    """Check for attribute."""
    if not value or not value.get_secret_value():
        error_msg = (
            f"Did not find {key}, please add an environment variable"
            f" `{env_key}` which contains it, or pass"
            f" `{key}` as a named parameter."
        )
        raise ValueError(error_msg)


def extract_params(
    kwargs: dict[str, Any],
    default_params: BaseSchema | dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Extract params."""
    if kwargs.get("params") is not None:
        params = kwargs.pop("params")
    elif default_params is not None:
        params = default_params
    else:
        params = None

    if isinstance(params, BaseSchema):
        params = params.to_dict()

    return params or {}


def extract_chat_params(
    kwargs: dict[str, Any],
    default_params: BaseSchema | dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Extract chat params."""
    if kwargs.get("params") is not None:
        params = kwargs.pop("params")
        check_duplicate_chat_params(params, kwargs)
    elif default_params is not None:
        params = deepcopy(default_params)
    else:
        params = None

    if isinstance(params, BaseSchema):
        params = params.to_dict()

    return params or {}


def check_duplicate_chat_params(params: dict[str, Any], kwargs: dict[str, Any]) -> None:
    """Check duplicate chat params."""
    duplicate_keys = {k for k, v in kwargs.items() if v is not None and k in params}

    if duplicate_keys:
        error_msg = (
            f"Duplicate parameters found in params and keyword arguments: "
            f"{list(duplicate_keys)}"
        )
        raise ValueError(error_msg)


def gateway_error_handler(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to catch ApiRequestFailure on Model Gateway calls.

    Logs a uniform warning when the model is not properly registered.
    """

    @functools.wraps(func)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        try:
            return func(self, *args, **kwargs)
        except ApiRequestFailure:
            if any(
                hasattr(self, attr)
                for attr in ("watsonx_model_gateway", "watsonx_embed_gateway")
            ):
                logger.warning(
                    "You are calling the Model Gateway endpoint using the 'model' "
                    "parameter. Please ensure this model is registered with the "
                    "Gateway. If you intend to use a watsonx.ai-hosted model, pass "
                    "'model_id' instead of 'model'.",
                )
            raise

    return wrapper


def async_gateway_error_handler(func: Callable[..., Any]) -> Callable[..., Any]:
    """Async decorator to catch ApiRequestFailure on Model Gateway calls.

    Log a uniform warning when the Model Gateway is misused.
    """

    @functools.wraps(func)
    async def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        try:
            return await func(self, *args, **kwargs)
        except ApiRequestFailure:
            if getattr(self, "watsonx_model_gateway", None) is not None:
                logger.warning(
                    "You are calling the Model Gateway endpoint using the 'model' "
                    "parameter. Please ensure this model is registered with the "
                    "Gateway. If you intend to use a watsonx.ai-hosted model, pass "
                    "'model_id' instead of 'model'.",
                )
            raise

    return wrapper


def resolve_watsonx_credentials(
    url: SecretStr,
    *,
    api_key: SecretStr | None = None,
    token: SecretStr | None = None,
    password: SecretStr | None = None,
    username: SecretStr | None = None,
    instance_id: SecretStr | None = None,
    version: SecretStr | None = None,
    verify: bool | str | None = None,
) -> Credentials:
    """Resolve watsonx credentials."""
    check_for_attribute(url, "url", "WATSONX_URL")

    raw_url = url.get_secret_value()
    parsed_url = urlparse(raw_url)
    clean_url = f"{parsed_url.scheme}://{parsed_url.hostname}"

    if clean_url in APIClient.PLATFORM_URLS_MAP:
        if not token and not api_key:
            error_msg = (
                "Did not find 'api_key' or 'token',"
                " please add an environment variable"
                " `WATSONX_API_KEY` or 'WATSONX_TOKEN' "
                "which contains it,"
                " or pass 'api_key' or 'token'"
                " as a named parameter."
            )
            raise ValueError(error_msg)
    elif not token and not password and not api_key:
        error_msg = (
            "Did not find 'token', 'password' or 'api_key',"
            " please add an environment variable"
            " `WATSONX_TOKEN`, 'WATSONX_PASSWORD' or 'WATSONX_API_KEY' "
            "which contains it,"
            " or pass 'token', 'password' or 'api_key'"
            " as a named parameter."
        )
        raise ValueError(error_msg)
    elif token:
        check_for_attribute(token, "token", "WATSONX_TOKEN")
    elif password:
        check_for_attribute(password, "password", "WATSONX_PASSWORD")
        check_for_attribute(username, "username", "WATSONX_USERNAME")
    elif api_key:
        check_for_attribute(api_key, "api_key", "WATSONX_API_KEY")
        check_for_attribute(username, "username", "WATSONX_USERNAME")

    return Credentials(
        url=url.get_secret_value() if url else None,
        api_key=api_key.get_secret_value() if api_key else None,
        token=token.get_secret_value() if token else None,
        password=password.get_secret_value() if password else None,
        username=username.get_secret_value() if username else None,
        instance_id=instance_id.get_secret_value() if instance_id else None,
        version=version.get_secret_value() if version else None,
        verify=verify,
    )


def secret_from_env_multi(
    names_priority: list[str], deprecated: set[str] | None = None
) -> Callable[[], SecretStr | None]:
    """Return default factory that yields a SecretStr from the first non-empty env var.

    The factory:
    - Warns if multiple environment variables are set
        (uses the first in `names_priority`).
    - Warns if the chosen environment variable is listed in `deprecated`.
    """
    deprecated = deprecated or set()

    def _factory() -> SecretStr | None:
        present = [(n, os.getenv(n)) for n in names_priority]
        present = [(n, v) for n, v in present if v not in (None, "")]
        if not present:
            return None

        chosen_name, value = present[0]
        if len(present) > 1:
            warnings.warn(
                f"Multiple API key env vars are set {[n for n, _ in present]}; "
                f"using '{chosen_name}'.",
                UserWarning,
                stacklevel=2,
            )
        if chosen_name in deprecated:
            warnings.warn(
                f"Environment variable '{chosen_name}' is deprecated; "
                f"use '{names_priority[0]}' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        return SecretStr(cast("str", value))

    return _factory


def normalize_api_key(data: dict[str, Any]) -> dict[str, Any]:
    """Normalize deprecated 'apikey' to 'api_key'.

    - If only 'apikey' is provided, convert it to 'api_key' with a DeprecationWarning.
    - If both are provided, 'api_key' takes precedence with a UserWarning.
    """
    has_new = "api_key" in data
    has_old = "apikey" in data

    if has_old and not has_new:
        warnings.warn(
            "'apikey' parameter is deprecated; use 'api_key' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        data = {**data, "api_key": data.pop("apikey")}
    elif has_old and has_new:
        warnings.warn(
            "Both 'api_key' and deprecated 'apikey' were provided; "
            "'api_key' takes precedence.",
            UserWarning,
            stacklevel=2,
        )
        data = {**data}
        data.pop("apikey", None)

    return data



from typing import Any

from ibm_watsonx_ai import APIClient, Credentials  # type: ignore[import-untyped]
from ibm_watsonx_ai.service_instance import (  # type: ignore[import-untyped]
    ServiceInstance,
)
from langchain_core.tools import BaseTool
from langchain_tests.unit_tests.tools import ToolsUnitTests

from langchain_ibm.agent_toolkits.utility.toolkit import WatsonxTool

client = APIClient.__new__(APIClient)
client.CLOUD_PLATFORM_SPACES = True
client.ICP_PLATFORM_SPACES = True
credentials = Credentials(api_key="api_key")
client.credentials = credentials
client.service_instance = ServiceInstance.__new__(ServiceInstance)


class TestWatsonxToolsStandard(ToolsUnitTests):
    @property
    def tool_constructor(self) -> type[BaseTool] | BaseTool:
        return WatsonxTool

    @property
    def tool_constructor_params(self) -> dict[str, Any]:
        return {
            "name": "GoogleSearch",
            "description": "Search for online trends, news, current events, "
            "real-time information, or research topics.",
            "agent_description": "Search for online trends, news, current events, "
            "real-time information, or research topics.",
            "tool_config_schema": {
                "title": "config schema for GoogleSearch tool",
                "type": "object",
                "properties": {
                    "maxResults": {
                        "title": "Max number of results to return",
                        "type": "integer",
                        "minimum": 1,
                        "maximum": 20,
                    }
                },
            },
            "watsonx_client": client,
        }

    @property
    def tool_invoke_params_example(self) -> dict[str, Any]:
        """
        Returns a dictionary representing the "args" of an example tool call.

        This should NOT be a ToolCall dict - i.e. it should not
        have {"name", "id", "args"} keys.
        """
        return {
            "input": "Search IBM",
        }
